import sys
import shutil

from . import utils

import leveldb


class kmerSetDB(object):
    '''
    Plyvel based scalable on disk kmerSetDB
    for k-mer storage.
    '''

    def __init__(self, path, homology, verbose=False):
        '''
        kmerSetDB constructor.
        '''
        try:
            # Assert path is string
            if not isinstance(path, str):
                print('\n[Non-Repetitive Parts Calculator - Background]')
                print('\n [ERROR]    Path must be a string, not {}'.format(
                    path))
                print(' [SOLUTION] Try correcting Path\n')
                raise ValueError

            # Assert homology is positive integer
            if not isinstance(homology, int):
                print('\n[Non-Repetitive Parts Calculator - Background]')
                print('\n [ERROR]    Lmax must be an integer, not {}'.format(type(homology)))
                print(' [SOLUTION] Try correcting Lmax\n')
                raise ValueError
            if homology-1 < 5:
                print('\n[Non-Repetitive Parts Calculator - Background]')
                print('\n [ERROR]    Lmax must be greater than 4, not \'{}\''.format(homology-1))
                print(' [SOLUTION] Try correcting Lmax\n')
                raise ValueError

            # Path setup
            self.PATH = path.rstrip('/') + '/'

            # Create/Open Plyvel object
            self.DB = leveldb.LevelDB(
                filename=self.PATH,
                create_if_missing=True,
                error_if_exists=False)

            # Length setup
            try:
                self.LEN = int(self.DB.Get(b'LEN'))
            except:
                self.LEN = 0
                self.DB.Put(b'LEN', b'0')

            # K setup
            try:
                self.K = int(self.DB.Get(b'K'))
            except:
                self.K = int(homology)
                self.DB.Put(b'K', str(homology).encode())

            # Verbosity setup
            self.VERB = bool(verbose)

            # Object ALIVE status
            self.ALIVE = True

        except ValueError as E:
            raise ValueError('Invalid Path or Lmax')
        except Exception as E:
            raise E

    def __repr__(self):
        '''
        User function to return the string representation of
        kmerSetDB.
        '''
        return 'kmerSetDB stored at {} with {} {}-mers'.format(
            self.PATH, self.LEN, self.K)

    def __str__(self):
        '''
        See __repr__
        '''
        return self.__repr__()

    def alivemethod(method):
        '''
        Internal decorator to gate kmerSetDB operation
        once dropped.
        '''
        def wrapper(self, *args, **kwargs):
            if self.ALIVE:
                return method(self, *args, **kwargs)
            else:
                raise RuntimeError('kmerSetDB was closed or dropped')
        return wrapper

    def _verb_action(self, action, index, seq):
        '''
        Internal helper function to print sequence
        updates to kmerSetDB based on verbosity.
        '''
        if self.VERB:
            psl = min(len(seq), 10)
            printed = '  {} Seq {}: {}...'.format(
                action, index, seq[:psl])
            clear_length = len(printed)
            sys.stdout.write(' '*clear_length+'\r')
            sys.stdout.write(printed)
            sys.stdout.flush()

    def _get(self, key):
        '''
        Internal helper to fetch value for given
        key, from kmerSetDB instance.
        '''
        try:
            val = self.DB.Get(key)
            return str(val)
        except:
            return None

    @alivemethod
    def add(self, seq):
        '''
        User function to add seq k-mers to kmerSetDB.
        '''
        try:
            if not seq in ['K', 'LEN']:
                seq = seq.replace('U', 'T')
                for kmer in utils.stream_min_kmers(
                    seq=seq, k=self.K):
                    kmer = kmer.encode()
                    if self._get(kmer) is None:
                        self.DB.Put(kmer, b'1')
                        self.LEN += 1
            self.DB.Put(b'LEN', str(self.LEN).encode())
        except Exception as E:
            raise E

    @alivemethod
    def multiadd(self, seq_list):
        '''
        User function to add seq k-mers for each seq in
        seq_list to kmerSetDB via single transaction.
        '''
        try:
            pt = False
            WB = leveldb.WriteBatch()
            index = 0
            for seq in seq_list:
                if not pt and self.VERB:
                    print('\n[Background Processing]')
                    pt = True
                if not seq in ['K', 'LEN']:
                    self._verb_action(
                        action='Adding',
                        index=index,
                        seq=seq)
                    index += 1
                    seq = seq.replace('U', 'T')
                    for kmer in utils.stream_min_kmers(
                        seq=seq, k=self.K):
                        kmer = kmer.encode()
                        if self._get(kmer) is None:
                            WB.Put(kmer, b'1')
                            self.LEN += 1
            WB.Put(b'LEN', str(self.LEN).encode())
            self.DB.Write(WB, sync=True)
            if self.VERB: print()
        except Exception as E:
            raise E

    @alivemethod
    def __contains__(self, seq, rna=True):
        '''
        Python dunder function to check existence of
        any k-mer from seq in kmerSetDB.
        '''
        try:
            if rna:
                seq = seq.replace('U', 'T')
            for kmer in utils.stream_min_kmers(
                seq=seq, k=self.K):
                kmer = kmer.encode()
                if self._get(kmer):
                    return True
            return False
        except Exception as E:
            raise E

    @alivemethod
    def multicheck(self, seq_list):
        '''
        User function to check existence of any k-mer for
        each seq in seq_list in kmerSetDB via single
        transaction.
        '''
        try:
            for seq in seq_list:
                seq = seq.replace('U', 'T')
                yield self.__contains__(
                    seq=seq,
                    rna=False)
        except Exception as E:
            raise E

    @alivemethod
    def __iter__(self):
        '''
        User fuction to iterate over k-mers stored in
        kmerSetDB.
        '''
        for key, _ in self.DB.RangeIter(
            key_from=None, key_to=None):
            if not key in [b'K', b'LEN']:
                yield str(key.decode('ascii'))

    @alivemethod
    def __len__(self):
        '''
        User function to return the number of keys stored
        in kmerSetDB.
        '''
        return self.LEN

    @alivemethod
    def remove(self, seq):
        '''
        User function to remove all k-mers in seq
        from kmerSetDB.
        '''
        try:
            if not seq in ['K', 'LEN']:
                seq = seq.replace('U', 'T')
                for kmer in utils.stream_min_kmers(
                    seq=seq, k=self.K):
                    kmer = kmer.encode()
                    if self._get(kmer):
                        self.DB.Delete(kmer)
                        self.LEN -= 1
                self.DB.Put(b'LEN', str(self.LEN).encode())
        except Exception as E:
            raise E

    @alivemethod
    def multiremove(self, seq_list, clear=False):
        '''
        User function to remove all k-mers from each
        seq in seq_list from kmerSetDB via single
        transaction.
        '''
        try:
            pt = False
            WB = leveldb.WriteBatch()
            index = 0
            for seq in seq_list:
                if not pt and self.VERB:
                    print('\n[Background Processing]')
                    pt = True
                if not seq in ['K', 'LEN']:
                    self._verb_action(
                        action='Removing',
                        index=index,
                        seq=seq)
                    index += 1
                    seq = seq.replace('U', 'T')
                    for kmer in utils.stream_min_kmers(
                        seq=seq, k=self.K):
                        kmer = kmer.encode()
                        if clear:
                            WB.Delete(kmer)
                            self.LEN -= 1
                        else:
                            if self._get(kmer):
                                WB.Delete(kmer)
                                self.LEN -= 1
            self.DB.Write(WB, sync=False)
            self.DB.Put(b'LEN', str(self.LEN).encode())
            if self.VERB: print()
        except Exception as E:
            raise E

    def clear(self, Lmax=None):
        '''
        User function to clear all k-mers stored in
        kmerSetDB.
        '''
        self.drop()
        self.DB = leveldb.LevelDB(
            filename=self.PATH,
            create_if_missing=True,
            error_if_exists=False)
        self.LEN = 0
        self.DB.Put(b'LEN', b'0')
        if Lmax is None:
            self.DB.Put(b'K', str(self.K).encode())
        else:
            if not Lmax is None:
                if not isinstance(Lmax, int):
                    print('\n [ERROR]    Lmax must be an integer, not {}'.format(
                        type(Lmax)))
                    print(' [SOLUTION] Try correcting Lmax\n')
                    raise ValueError
                self.K = int(Lmax) + 1
                if Lmax < 5:
                    print('\n [ERROR]    Lmax must be greater than 4, not {}'.format(
                        Lmax))
                    print(' [SOLUTION] Try correcting Lmax\n')
                    raise ValueError
            self.DB.Put(b'K', str(self.K).encode())
        self.ALIVE = True

    def close(self):
        '''
        User function to close kmerSetDB.
        '''
        if self.ALIVE:
            del self.DB
            self.DB = None
            self.ALIVE = False
            return True
        return False

    def drop(self):
        '''
        User function to drop kmerSetDB.
        '''
        if self.ALIVE:
            del self.DB
            self.DB = None
            shutil.rmtree(self.PATH)
            self.ALIVE = False
            return True
        return False

def test():

    # required test imports
    import random
    import time

    # returns a random DNA string of length
    def get_DNA(length):
        return ''.join(random.choice('ATGC') for _ in range(length))

    # create corpus
    total_elements = 10000
    dna_strings = sorted(get_DNA(100) for _ in range(total_elements))
    print('Testing Operations on {} items'.format(total_elements*2))

    # clearance
    try:
        shutil.rmtree('./testBG/')
    except:
        pass

    # object test
    myBG = kmerSetDB(homology=99, path='./testBG')
    assert isinstance(myBG, kmerSetDB)
    assert repr(myBG) == 'kmerSetDB stored at ./testBG/ with 0 99-mers'

    # add test
    t0 = time.time()
    for i in range(total_elements):
        myBG.add(dna_strings[i])
    print('add          time = {} sec'.format(time.time()-t0))
    myBG.clear()

    # # multiadd test
    t0 = time.time()
    myBG.multiadd(dna_strings)
    print('multiadd     time = {} sec'.format(time.time()-t0))

    # __contains__ test
    assert get_DNA(length=100) not in myBG
    t0 = time.time()
    for dna in dna_strings:
        assert dna in myBG
    print('__contains__ time = {} sec'.format(time.time()-t0))

    # multicheck test
    t0 = time.time()
    assert all(myBG.multicheck(seq_list=dna_strings))
    print('multicheck   time = {} sec'.format(time.time()-t0))

    # __iter__ test
    t0 = time.time()
    sorted(myBG)
    print('__iter__     time = {} sec'.format(time.time()-t0))

    # __len__ test
    t0 = time.time()
    assert len(myBG) == total_elements * 2
    print('__len__      time = {} sec'.format(time.time()-t0))

    # remove test
    t0 = time.time()
    assert len(myBG) == total_elements*2
    for dna_string in dna_strings:
        myBG.remove(dna_string)
    print('remove       time = {} sec'.format(time.time()-t0))
    assert len(myBG) == 0

    # multiremove test
    myBG.multiadd(dna_strings)
    assert len(myBG) == total_elements*2
    t0 = time.time()
    myBG.multiremove(dna_strings)
    print('multiremove  time = {} sec'.format(time.time()-t0))
    assert len(myBG) == 0

    # clear test
    myBG.multiadd(dna_strings)
    t0 = time.time()
    myBG.clear()
    print('clear        time = {} sec'.format(time.time()-t0))

    # close test
    t0 = time.time()
    myBG.close()
    print('close        time = {} sec'.format(time.time()-t0))

    # drop test
    myBG = kmerSetDB(homology=99, path='./testBG')
    t0 = time.time()
    myBG.drop()
    print('drop         time = {} sec'.format(time.time()-t0))

if __name__ == '__main__':
    test()


import random
import functools
import collections

iupac_count = {
    'A': 1, 'C': 1, 'G': 1, 'T': 1,
    'R': 2, 'Y': 2, 'S': 2, 'W': 2,
    'K': 2, 'M': 2, 'B': 3, 'D': 3,
    'H': 3, 'V': 3, 'N': 4, 'U': 1
}

iupac_space_rna = {
    'A': {'A'}, 'C': {'C'}, 'G': {'G'}, 'U': {'U'},
    'R': {'A', 'G'}, 'Y': {'C', 'U'},
    'S': {'G', 'C'}, 'W': {'A', 'U'},
    'K': {'G', 'U'}, 'M': {'A', 'C'},
    'B': {'C', 'G', 'U'}, 'V': {'A', 'C', 'G'},
    'D': {'A', 'G', 'U'}, 'H': {'A', 'C', 'U'},
    'N': {'A', 'U', 'G', 'C'}
}

iupac_compl_rna = {
    'A': 'U', 'C': 'G', 'G': 'Y', 'U': 'R',
    'R': 'Y', 'Y': 'R', 'S': 'B', 'W': 'D',
    'K': 'N', 'M': 'K', 'B': 'N', 'V': 'D',
    'D': 'N', 'H': 'N', 'N': 'N'
}

iupac_space_dna = {
    'A': {'A'}, 'C': {'C'}, 'G': {'G'}, 'T': {'T'},
    'R': {'A', 'G'}, 'Y': {'C', 'T'},
    'S': {'G', 'C'}, 'W': {'A', 'T'},
    'K': {'G', 'T'}, 'M': {'A', 'C'},
    'B': {'C', 'G', 'T'}, 'V': {'A', 'C', 'G'},
    'D': {'A', 'G', 'T'}, 'H': {'A', 'C', 'T'},
    'N': {'A', 'T', 'G', 'C'}
}

iupac_compl_dna = {
    'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A',
    'R': 'Y', 'Y': 'R', 'S': 'S', 'W': 'W',
    'K': 'M', 'M': 'K', 'B': 'V', 'V': 'B',
    'D': 'H', 'H': 'D', 'N': 'N'
}

def is_homolog_legal(seq, homology):
    '''
    Check is homology is satisfiable
    '''
    if homology > len(seq):
        return False
    return True

def is_seq_constr_legal(seq, part_type):
    '''
    Check if all characters used is standard IUPAC code.
    On failure returns (False, a list of illegal chars used).
    '''
    global iupac_space_dna
    global iupac_space_rna
    if part_type == 'DNA':
        iupac_space = iupac_space_dna
    else:
        iupac_space = iupac_space_rna
    chars    = set(seq)
    alphabet = set(iupac_space.keys())
    if chars <= alphabet:
        return (True, None)
    else:
        return (False, sorted(chars-alphabet))

def get_computable_form(struct):
    '''
    Get computable components from an RNA secondary structure.
    '''
    opened  = []
    closed  = []
    pairs   = []
    invalid = []
    for pos in range(len(struct)):
        pairing = struct[pos]
        if pairing == '(':
            opened.append(pos)
        elif pairing == ')':
            closed.append(pos)
            try:
                pairs.append((opened.pop(), closed.pop()))
            except:
                pass
        elif pairing not in ['x', '.']:
            invalid.append(pos)
    return pairs, opened, closed, invalid

def get_k_mer_count(seq_list, start, end, rev_dict):
    '''
    Get the possible number of candidate sequences.
    '''
    global iupac_count
    product = 1
    for j in range(start, end):
        nt = seq_list[j]
        if j in rev_dict:
            i = rev_dict[j]
            if i >= start:
                continue
        product *= iupac_count[nt]
    return product

def compress_locs(locs, homology):
    '''
    Compress contigs in locs.
    '''
    if not locs:
        return None
    else:
        compressed = []
        x = locs[0]
        y = x+homology
        for pos in range(1, len(locs)):
            if locs[pos]-locs[pos-1] == 1:
                y = locs[pos]+homology
            else:
                compressed.append((x, y))
                x = locs[pos]
                y = x+homology
        compressed.append((x, y))
    return compressed

def is_seq_constr_sufficient(seq, struct, homology, toolbox_size):
    '''
    Check if a desired size toolbox can be generated.
    On failure returns (False, a list of start:end tuples with constricted motifs).
    '''
    seq_list = list(seq)
    sufficiency_status = True
    insufficiency_locs = []
    rev_dict = {j:i for i,j in get_computable_form(struct)[0]}
    for i in range(len(seq)-homology+1):
        start = i
        end = start + homology
        kmer_count = get_k_mer_count(
            seq_list=seq_list,
            start=start,
            end=end,
            rev_dict=rev_dict)
        if kmer_count < toolbox_size:
            sufficiency_status = False
            insufficiency_locs.append(i)
    return (sufficiency_status, compress_locs(locs=insufficiency_locs, homology=homology))

def is_structure_valid(struct):
    '''
    Check if the given RNA structure is legal and balanced.
    On failure returns (False, three lists of indices:
    unclosed, unopened and invalid charas/parens)
    '''
    pairs, opened, closed, invalid = get_computable_form(struct)
    if not opened:
        opened  = None
    if not closed:
        closed  = None
    if not invalid:
        invalid = None
    if opened or closed or invalid:
        return (False, opened, closed, invalid)
    else:
        return (True, None, None, None)

def hamming(a, b):
    '''
    Get hamming distance between two strings.
    '''
    hdist = 0
    for i in range(len(a)):
        if a[i] != b[i]:
            hdist += 1
    return hdist

def is_contiguous(item1, item2):
    '''
    Check if two pairs are contiguous.
    '''
    if (item1[0]-item2[0] <= 1) and (item2[1]-item1[1] <= 1):
        return True
    return False

def is_structure_not_conflict(struct, homology):
    '''
    Q. Does the structure not enforce internal repeats?
    A. Stack evalation of stack of pairs.
    On failure returns (False, a list of locations with stems >= homology)
    '''
    pairs, opened, closed, invalid = get_computable_form(struct)
    if pairs:
        bad_contigs = []
        stretch     = 1
        pairs       = pairs[::-1]
        start       = pairs.pop()
        end         = None
        item1       = start
        item2       = end
        while pairs:
            end   = pairs.pop()
            item2 = end
            if is_contiguous(item1, item2):
                item1 = item2
                item2 = None
                stretch += 1
            else:
                if stretch >= homology:
                    bad_contigs.append(((item1[0], start[0]), (start[1], item1[1])))
                stretch = 1
                start = item2
                item1 = start
        if stretch >= homology:
            bad_contigs.append(((end[0], start[0]), (start[1], end[1])))
        if bad_contigs:
            return (False, bad_contigs)
    return (True, None)

def is_pairing_compatible(seq, struct, part_type):
    global iupac_space_dna
    global iupac_compl_dna
    global iupac_space_rna
    global iupac_compl_rna
    pairs, opened, closed, invalid = get_computable_form(struct)
    incompat_locs = []
    reduced_locs = []
    if pairs:
        if part_type == 'DNA':
            iupac_space = iupac_space_dna
            iupac_compl = iupac_compl_dna
        else:
            iupac_space = iupac_space_rna
            iupac_compl = iupac_compl_rna
        while pairs:
            i,j = pairs.pop()
            nti = seq[i]
            ntj = seq[j]
            if len(iupac_space[nti]) > len(iupac_space[ntj]):
                nti,ntj = ntj,nti
            if not iupac_space[nti] <= iupac_space[iupac_compl[ntj]]:
                incompat_locs.append(tuple(sorted([i,j])))
            else:
                if iupac_space[iupac_compl[ntj]] > iupac_space[nti]:
                    reduced_locs.append(tuple(sorted([i,j])))
        if len(incompat_locs) > 0 or \
           len(reduced_locs) > 0:
            return (False, incompat_locs, reduced_locs)
    return (True, incompat_locs, reduced_locs)

def main():
    print('Testing Sequence Constraint Validation Checkers ... ',)
    legal_alphabet  = set('ACGTRYSWKMBDHVN')
    seq_legal = ''.join(random.sample(legal_alphabet, 1)[0] for _ in range(100))
    assert is_seq_constr_legal(seq=seq_legal, part_type='DNA')[0] == True

    illegal_alphabet = set('ABCDEFGHIJKLMNOPQRSTUVWXYZ')
    seq_illegal = ''.join(random.sample(illegal_alphabet, 1)[0] for _ in range(100))
    assert is_seq_constr_legal(seq=seq_illegal, part_type='DNA')[0] == False
    print('OK')

    print('Testing Sequence Constraint Sufficiency Checkers ... ',)
    seq      = 'N'*20 + 'TTGACA' + 'N'*17 + 'TATAAT' + 'N'*6 + 'CCN' + 'N'*20
    struct   = '.'*len(seq)
    homology = 11
    assert is_seq_constr_sufficient(seq, struct, homology, toolbox_size=1000)[0] == True
    assert is_seq_constr_sufficient(seq, struct, homology, toolbox_size=2000)[0] == False
    struct = '......((((((((((xxxxxxxx))))))))))......'
    seq    = 'NNNNNNNNNNNNNNNNAAAAAAAANNNNNNNNNNNNNNNN'
    assert is_seq_constr_sufficient(seq, struct, homology, toolbox_size=1000)[0] == False
    assert is_seq_constr_sufficient(seq, struct, 17, toolbox_size=1000)[0] == True
    print('OK')

    print('Testing Struture Constraint Validation Checkers ...',)
    struct = '....(((.xx.)))...()...(())...'
    assert is_structure_valid(struct)[0] == True
    struct = '....(((.xx.))).).()...(())...'
    assert is_structure_valid(struct)[0] == False
    struct = '.(..(((.xx.)))...()...(())...'
    assert is_structure_valid(struct)[0] == False
    struct = '....(((.xx.))).X.()...(())...'
    assert is_structure_valid(struct)[0] == False
    struct = '.B..(((.xx.))).).().(.(())...'
    assert is_structure_valid(struct)[0] == False
    print('OK')

    print('Testing Struture Constraint Homology Conflict Checkers ...',)
    struct = '(((((((((((......)))))))))))...(((((((((((......)))))))))))'
    homology = 10
    assert is_structure_not_conflict(struct, homology)[0] == False
    struct = '(((((((((((......))))).))))))...(((((((((((......)))))))))))'
    homology = 10
    assert is_structure_not_conflict(struct, homology)[0] == False
    struct = '(((((((((((......)))).)))))))...(((((.((((((......)))))))))))'
    homology = 10
    assert is_structure_not_conflict(struct, homology)[0] == True
    print('OK')

    print('Testing Sequence Constraint Base Pairing Conflict Checkers ...',)
    seq    = 'NNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNN'
    struct = '.....(((..(((((....)))))..))).....'
    assert is_pairing_compatible(seq, struct, part_type='DNA') == (True, [], [])
    seq    = 'NNNNNNNNNNSNNNNNNNNNNNNSNNNNNNNNNN'
    struct = '.....(((..(((((....)))))..))).....'
    assert is_pairing_compatible(seq, struct, part_type='DNA') == (True, [], [])
    seq    = 'NNNNNNNNNNSNNNNNNNNNNNNWNNNNNNNNNN'
    struct = '.....(((..(((((....)))))..))).....'
    assert is_pairing_compatible(seq, struct, part_type='DNA') == (False, [(10, 23)], [])
    seq    = 'NNNNNNNNNNNNNNNNNNNNNNNANNNNNNNNNN'
    struct = '.....(((..(((((....)))))..))).....'
    assert is_pairing_compatible(seq, struct, part_type='RNA') == (False, [], [(10, 23)])
    seq    = 'NNNNNANNNNNNNNNNNNNNNNNNNNNNNNNNNN'
    struct = '.....(((..(((((....)))))..))).....'
    assert is_pairing_compatible(seq, struct, part_type='RNA') == (False, [], [(5, 28)])
    seq    = 'NNNNNMNNNNNNNNNNNNNNNNNNNNNNVNNNNN'
    struct = '.....(((..(((((....)))))..))).....'
    assert is_pairing_compatible(seq, struct, part_type='RNA') == (False, [(5, 28)], [])
    seq    = 'NNNNNHNNNNNNNNNNNNNNNNNNNNNNCNNNNN'
    struct = '.....(((..(((((....)))))..))).....'
    assert is_pairing_compatible(seq, struct, part_type='RNA') == (False, [], [(5, 28)])
    seq    = 'NNNNNGNNNNNNNNNNNNNNNNNNNNNNTNNNNN'
    struct = '.....(((..(((((....)))))..))).....'
    assert is_pairing_compatible(seq, struct, part_type='DNA') == (False, [(5, 28)], [])
    seq    = 'NNNNNGNNNNNNNNNNNNNNNNNNNNNNUNNNNN'
    struct = '.....(((..(((((....)))))..))).....'
    assert is_pairing_compatible(seq, struct, part_type='RNA') == (False, [], [(5, 28)])
    print('OK')

if __name__ == '__main__':
    main()




class kmerSetArray(object):
    '''
    Custom set-array structure for
    k-mer storage.
    '''

    def __init__(self, size):
        '''
        kmerSetArray constructor.
        '''
        try:
            self.unique_kmers = dict()
            self.indexd_kmers = [' '] * size
        except Exception as E:
            raise E

    def __setitem__(self, index, kmer):
        '''
        User function to set kmer at given index.
        '''
        try:
            if not self.indexd_kmers[index] == ' ':
                if self.indexd_kmers[index] in self.unique_kmers:
                    self.unique_kmers.pop(
                        self.indexd_kmers[index])
            self.unique_kmers[kmer]  = index
            self.indexd_kmers[index] = kmer
        except Exception as E:
            raise E

    def __contains__(self, kmer):
        '''
        User function to check existence of kmer
        in kmerSetArray.
        '''
        if kmer in self.unique_kmers:
            return True
        else:
            return False

    def index(self, kmer):
        '''
        User function to return the index of
        kmer if stored in kmerSetArray.
        '''
        if kmer in self.unique_kmers:
            return self.unique_kmers[kmer]
        return None

    def __len__(self):
        '''
        User function to return the number of
        kmers stored in kmerSetArray.
        '''
        return len(self.unique_kmers)

    def __repr__(self):
        '''
        User function to return a string
        representation of kmerSetArray.
        '''
        return 'kmerSetArray[{}]'.format(
            ', '.join(map(str, self.indexd_kmers)))

    def __str__(self):
        '''
        See __repr__
        '''
        return self.__repr__()

    def __del__(self):
        '''
        kmerSetArray destructor.
        '''
        self.unique_kmers = None
        self.indexd_kmers = None

def test():

    # required test imports
    import random
    import time

    from . import utils

    # returns a random DNA string of length
    def get_DNA(length):
        return ''.join(random.choice('ATGC') for _ in xrange(length))

    # create seq
    seq = get_DNA(length=100)
    print('Testing Operations on {}...'.format(seq[:25], seq[-25:]))

    # object test
    mySet = kmerSetArray(size=len(seq)-16+1)

    # __setitem__ test
    t0 = time.time()
    for i,kmer in enumerate(utils.stream_kmers(seq, k=16)):
        mySet[i] = kmer
    print('__setitem__  time = {} sec'.format(time.time()-t0))

    # __contains__ test
    t0 = time.time()
    for i,kmer in enumerate(utils.stream_kmers(seq, k=16)):
        assert kmer in mySet
    print('__contains__ time = {} sec'.format(time.time()-t0))

    # index test
    t0 = time.time()
    for i,kmer in enumerate(utils.stream_kmers(seq, k=16)):
        assert mySet.index(kmer) == i
    print('index        time = {} sec'.format(time.time()-t0))

    # __len__ test
    t0 = time.time()
    assert len(mySet) == len(seq)-16+1
    print('__len__      time = {} sec'.format(time.time()-t0))

    # __del__ test
    t0 = time.time()
    del mySet
    print('__del__      time = {} sec'.format(time.time()-t0))


if __name__ == '__main__':
    test()


from . import hgraph
from . import recovery
from . import utils
from . import kmerSetDB
from . import projector

import os
import uuid
import atexit
import shutil
import textwrap

import time

def _check_finder_constraints(
    seq_list,
    homology,
    allow_internal_repeat,
    vercov_func):
    # Sequence List Legality
    for seq in seq_list:
        if not isinstance(seq, str):
            print(' [ERROR]    Parts in Sequence List must be string, not {}'.format(type(seq)))
            print(' [SOLUTION] Try correcting Lmax\n')
            return False
    # Lmax Legality 1
    if not isinstance(homology, int):
        print('\n [ERROR]    Lmax must be an integer, not {}'.format(type(homology)))
        print(' [SOLUTION] Try correcting Lmax\n')
        return False
    # Lmax Legality 2
    if homology-1 < 5:
        print('\n [ERROR]    Lmax must be greater than 4, not {}'.format(homology-1))
        print(' [SOLUTION] Try correcting Lmax\n')
        return False
    # Internal Repeats Legality
    if not allow_internal_repeat in [True, False]:
        print('\n [ERROR]    Internal Repeat must be boolean, not {}'.format(
            type(allow_internal_repeat)))
        print(' [SOLUTION] Try correcting Internal Repeat\n')
        return False
    # Everything OK
    return True

def _check_finder_arguments(
    vercov_func,
    output_file):
    # Vertex Cover Legality 1
    if not isinstance(vercov_func, str):
        print(' [ERROR]    Vertex Cover Routine Name must be string, not {}'.format(type(vercov_func)))
        print(' [SOLUTION] Try correcting Vertex Cover Routine Name\n')
        return False
    # Vertex Cover Legality 2
    if not vercov_func in ['2apx', 'nrpG', 'nrp2']:
        print('\n [ERROR]    Vertex Cover Routine Name must be \'2apx\', \'nrpG\', or \'nrp2\', not \'{}\''.format(
            vercov_func))
        print(' [SOLUTION] Try correcting Vertex Cover Routine Name\n')
        return False
    # Output File Legality
    if not output_file is None:
        if not isinstance(output_file, str):
            print('\n [ERROR]    Output File must be a string or None, not {}'.format(type(output_file)))
            print(' [SOLUTION] Try correcting Output File\n')
            return False
    # Everything OK
    return True

def nrp_finder(
    seq_list,
    homology,
    allow_internal_repeat=False,
    background=None,
    vercov_func=None,
    output_file=None,
    verbose=True,
    check_constraints=True):

    # Program Verbage
    if verbose:
        print('\n[Non-Repetitive Parts Calculator - Finder Mode]')
    find_parts = True
    seq_list = list(seq_list)

    # Check Finder Constraints
    if check_constraints:
        if verbose:
            print('\n[Checking Constraints]')
            print(' Sequence List   : {} parts'.format(len(seq_list)))
            print('          Lmax   : {} bp'.format(homology-1))
            print(' Internal Repeats: {}'.format(allow_internal_repeat))
        check_status = _check_finder_constraints(
            seq_list=seq_list,
            homology=homology,
            allow_internal_repeat=allow_internal_repeat,
            vercov_func=vercov_func)

        if check_status == False:
            if verbose:
                print(' Check Status: FAIL\n')
            find_parts = False
        else:
            if verbose:
                print('\n Check Status: PASS')

        # Background Check
        if find_parts:
            if not background is None:
                if verbose:
                    print('\n[Checking Background]\n Background: {}'.format(background))
                if isinstance(background, kmerSetDB.kmerSetDB):
                    if not background.ALIVE:
                        build_parts = False
                        print('\n [ERROR]    Background is closed or dropped')
                        print(' [SOLUTION] Try using an open Background\n')
                        if verbose:
                            print(' Check Status: FAIL\n')
                    else:
                        if verbose:
                            print('\n Check Status: PASS')
                else:
                    find_parts = False
                    print('\n [ERROR]    Background Object is INVALID')
                    print(' [SOLUTION] Try instantiating background via nrpcalc.background(...)\n')
                    if verbose:
                        print(' Check Status : FAIL\n')

        # Arguments Check
        if find_parts:
            if verbose:
                print('\n[Checking Arguments]')
                print('   Vertex Cover: {}'.format(vercov_func))
                print('   Output  File: {}'.format(output_file))
            check_status = _check_finder_arguments(
                vercov_func=vercov_func,
                output_file=output_file)

            if check_status == False:
                if verbose:
                    print(' Check Status: FAIL\n')
                find_parts = False
            else:
                if verbose:
                    print('\n Check Status: PASS')

        if not find_parts:
            raise RuntimeError('Invalid Constraints, or Background')

    # Separate Checks from Discovery Logs
    if verbose:
        print()

    # Setup Project
    current_uuid = str(uuid.uuid4())
    projector.setup_proj_dir(current_uuid)
    seq_file   = './{}/seq_list.txt'.format(current_uuid)
    graph_file = './{}/repeat_graph.txt'.format(current_uuid)

    # Build Repeat Graph
    homology_graph = hgraph.get_homology_graph(
        seq_list,
        background,
        homology,
        allow_internal_repeat,
        seq_file,
        verbose)

    # Find Non-Repetitive Subset
    indi_seqs_indx = recovery.get_recovered_non_homologs(
        homology_graph,
        graph_file,
        vercov_func,
        verbose)

    # Prepare Output
    parts_dict = {}
    with open(seq_file, 'r') as infile:
        for line in infile:
            index,seq = line.strip().split(',')
            index = int(index)
            if index in indi_seqs_indx:
                parts_dict[index] = seq

    # Optionally, write output
    if output_file:
        with open(output_file, 'w') as outfile:
            for index,seq in parts_dict.items():
                outfile.write('>index {}\n'.format(index))
                seq_wrap = '\n'.join(textwrap.wrap(seq, 80))
                outfile.write('{}\n'.format(seq_wrap))

    # Cleanups and Return
    projector.remove_proj_dir()
    if verbose:
        print('\nNon-Repetitive Toolbox Size: {}'.format(
            len(parts_dict)))
    return parts_dict


import atexit
import os
import shutil

current_uuid = None

def setup_proj_dir(uuid):
    global current_uuid
    current_uuid = uuid
    directory = './{}/'.format(current_uuid)
    if not os.path.exists(directory):
        os.makedirs(directory)

@atexit.register
def remove_proj_dir():
    global current_uuid
    directory = './{}/'.format(current_uuid)
    if not current_uuid is None and os.path.isdir(directory):
        shutil.rmtree(directory)


import sys

import networkx as nx
import networkx.algorithms.approximation as nxa

from collections import deque
from random      import choice, shuffle

def nx_vercov(homology_graph, verbose):
    return nxa.vertex_cover.min_weighted_vertex_cover(homology_graph)

def get_max_degree_candidate_nodes(homology_graph, candidate_nodes):
    max_degree, max_degree_nodes = 0, []
    for node, degree in candidate_nodes:
        if degree == max_degree:
            max_degree_nodes.append(node)
        elif degree > max_degree:
            max_degree, max_degree_nodes = degree, [node]
    shuffle(max_degree_nodes)
    return max_degree_nodes

def get_max_degree_nodes(homology_graph, candidate_nodes=None):
    if candidate_nodes is None:
        return get_max_degree_candidate_nodes(homology_graph, candidate_nodes=homology_graph.degree())
    else:
        return get_max_degree_candidate_nodes(homology_graph, candidate_nodes=homology_graph.degree(candidate_nodes))

def get_pendants_from_node(homology_graph, candidate_node, verbose):
    pendant_list = []
    for neighbor in homology_graph[candidate_node]:
        if len(homology_graph[neighbor]) <= 2:
            pendant_list.append(neighbor)
    pendant_list.sort(key=lambda node: len(homology_graph[node])-1)
    return deque(pendant_list)

def remove_node_and_pendants(homology_graph, candidate_node, vercov_nodes, verbose):
    if candidate_node in homology_graph:
        pendant_deque = get_pendants_from_node(homology_graph, candidate_node, verbose)
        homology_graph.remove_node(candidate_node)
        if pendant_deque:
            eliminate_pendants(homology_graph, pendant_deque, vercov_nodes, verbose)

def remove_edge_and_pendants(homology_graph, candidate_edge, vercov_nodes, verbose):
    for node in candidate_edge:
        remove_node_and_pendants(homology_graph, node, vercov_nodes, verbose)

def eliminate_pendants(homology_graph, pendant_deque, vercov_nodes, verbose):
    while pendant_deque:
        candidate_node = pendant_deque.popleft()
        if candidate_node in homology_graph:
            if len(homology_graph[candidate_node]) == 1:
                if verbose:
                    print('  [x] Pendant node {} eliminated'.format(candidate_node))
                adj_node = next(iter(homology_graph[candidate_node]))
                homology_graph.remove_edge(adj_node, candidate_node)
                vercov_nodes.add(adj_node)
                new_candidates = [node for node in homology_graph[adj_node] if len(homology_graph[node]) <= 2]
                if new_candidates:
                    new_candidates.sort(key=lambda x: len(homology_graph[x])-1)
                    pendant_deque.extendleft(new_candidates)
                homology_graph.remove_nodes_from([candidate_node, adj_node])
            elif len(homology_graph[candidate_node]) == 0:
                if verbose:
                    print('  [x] Isolated node {} eliminated'.format(candidate_node))
                    homology_graph.remove_node(candidate_node)
        else:
            if verbose:
                print('  [+] Node {} covered'.format(candidate_node))

def get_pendant_deque(homology_graph):
    return deque(sorted([node for node in homology_graph if len(homology_graph[node]) <= 1], key=lambda x: len(homology_graph[x])))

def init_vercov(homology_graph, verbose):
    vercov_nodes = set()

    if verbose:
        print('\n Pendant checking is in progress...')
    pendant_deque = get_pendant_deque(homology_graph)

    if not pendant_deque:
        if verbose:
            print('  [x] No pendants found')
    else:
        if verbose:
            print('  [+] {} Pendants found\n\n Pendant elimination initiated...'.format(len(pendant_deque)))
        eliminate_pendants(homology_graph, pendant_deque, vercov_nodes, verbose)
        if verbose:
            print(    )
    return vercov_nodes

def get_arbitrary_neighbor(homology_graph, candidate_node, verbose):
    if len(homology_graph[candidate_node]) > 0:
        return homology_graph[candidate_node].iterkeys().next()
    return None

def get_sorted_nodes(homology_graph, verbose):
    return sorted(homology_graph.nodes(), key=lambda node: len(homology_graph[node]))

def std_vercov_approx(homology_graph, verbose):
    vercov_nodes = init_vercov(homology_graph, verbose)

    sorted_nodes = get_sorted_nodes(homology_graph, verbose)

    while homology_graph.number_of_edges():
        if verbose:
            print(' [Edges remaining = {}] [Nodes remaining = {}]'.format(homology_graph.number_of_edges(), homology_graph.number_of_nodes()))

        node_i = sorted_nodes.pop()
        if node_i in homology_graph:
            node_j = get_arbitrary_neighbor(homology_graph, node_i, verbose)
            if not node_j is None:
                vercov_nodes.update([node_i, node_j])
                remove_edge_and_pendants(homology_graph, candidate_edge=(node_i, node_j), vercov_nodes=vercov_nodes, verbose=verbose)
            else:
                homology_graph.remove_node(node_i)

    return vercov_nodes

def nrp_vercov_approx(homology_graph, verbose):
    vercov_nodes = init_vercov(homology_graph, verbose)

    while homology_graph.number_of_edges():
        if verbose:
            print(' [Edges remaining = {}] [Nodes remaining = {}]'.format(homology_graph.number_of_edges(), homology_graph.number_of_nodes()))

        max_degree_nodes = get_max_degree_nodes(homology_graph)

        for max_node in max_degree_nodes:
            if max_node in homology_graph:
                max_neighbor = get_max_degree_nodes(homology_graph, candidate_nodes=homology_graph.neighbors(max_node))
                if max_neighbor:
                    max_neighbor = max_neighbor.pop()
                    vercov_nodes.update([max_node, max_neighbor])
                    remove_edge_and_pendants(homology_graph, candidate_edge=(max_node, max_neighbor), vercov_nodes=vercov_nodes, verbose=verbose)
                else:
                    homology_graph.remove_node(max_node)

    return vercov_nodes

def nrp_vercov_greedy(homology_graph, verbose):
    vercov_nodes = init_vercov(homology_graph, verbose)

    while homology_graph.number_of_edges():
        if verbose:
            print(' [Edges remaining = {}] [Nodes remaining = {}]'.format(homology_graph.number_of_edges(), homology_graph.number_of_nodes()))

        max_degree_nodes = get_max_degree_nodes(homology_graph)

        for max_node in max_degree_nodes:
            if max_node in homology_graph:
                vercov_nodes.add(max_node)
                remove_node_and_pendants(homology_graph, candidate_node=max_node, vercov_nodes=vercov_nodes, verbose=verbose)

    return vercov_nodes

def test():
    print('Linear Chain')
    G = nx.Graph()
    for i in xrange(1, 10):
        G.add_edge(i, i+1)
    print('Nodes:', G.nodes())
    print('Edges:', G.edges())
    for i in xrange(1, 11):
        print('  {} -> {}'.format(i, list(G[i])))
    mvc = nrp_vercov_approx(G, verbose=True)
    print('Vertex Cover:', sorted(mvc))


if __name__ == '__main__':
    test()



from collections  import namedtuple
from itertools    import count
from glob         import glob
from time         import time, sleep

from .synthesis    import Synthesis
from .utils        import Fold
from .kmerSetDB    import kmerSetDB
from .kmerSetArray import kmerSetArray

import sys
import numpy
import uuid
import math
import textwrap

from . import utils
from . import makerchecks
from . import projector
from . import berno
from . import finder

from Bio.SeqUtils import MeltingTemp


class NRPMaker(object):

    def __init__(self, part_type='RNA', seed=None):

        # Instance variables
        self.part_type  = part_type
        self.synthesis  = Synthesis()
        self.fold       = Fold(part_type=part_type)
        self.proj_id    = str(uuid.uuid4())
        self.kmer_db    = None
        self.background = None

        # Seed the RNG
        if not seed is None and isinstance(seed, int):
            self.rng = numpy.random.default_rng(seed=seed)
        else:
            self.rng = numpy.random.default_rng()

        # Lookup tables
        if self.part_type == 'RNA':
            self.iupac_space = {
                'A': {'A'},
                'C': {'C'},
                'G': {'G'},
                'U': {'U'},
                'R': {'A', 'G'},
                'Y': {'C', 'U'},
                'S': {'G', 'C'},
                'W': {'A', 'U'},
                'K': {'G', 'U'},
                'M': {'A', 'C'},
                'B': {'C', 'G', 'U'},
                'V': {'A', 'C', 'G'},
                'D': {'A', 'G', 'U'},
                'H': {'A', 'C', 'U'},
                'N': {'A', 'U', 'G', 'C'}
            }
            self.iupac_compl = {
                'A': 'U', # A - U
                'C': 'G', # C - G
                'G': 'Y', # G - C U - Y
                'U': 'R', # U - A G - R
                'R': 'Y', # R - A G - U C - Y
                'Y': 'R', # Y - C U - G A G - R
                'S': 'B', # S - G C - C G U - B
                'W': 'D', # W - A U - U A G - D
                'K': 'N', # K - G U - C U G A
                'M': 'K', # M - A C - U G - K
                'B': 'N', # B - C G U - G C U A - N
                'V': 'D', # V - A C G - U G A - D
                'D': 'N', # D - A G U - U C A G - N
                'H': 'N', # H - A C U - U G A G - N
                'N': 'N'
            }
            self.base_compl = {
                'A': ['U'],
                'G': ['C', 'U'],
                'C': ['G'],
                'U': ['A', 'G']
            }
        else:
            self.iupac_space = {
                'A': {'A'},
                'C': {'C'},
                'G': {'G'},
                'T': {'T'},
                'R': {'A', 'G'},
                'Y': {'C', 'T'},
                'S': {'G', 'C'},
                'W': {'A', 'T'},
                'K': {'G', 'T'},
                'M': {'A', 'C'},
                'B': {'C', 'G', 'T'},
                'V': {'A', 'C', 'G'},
                'D': {'A', 'G', 'T'},
                'H': {'A', 'C', 'T'},
                'N': {'A', 'T', 'G', 'C'}
            }
            self.iupac_compl = {
                'A': 'T', # A - T
                'C': 'G', # C - G
                'G': 'C', # G - C
                'T': 'A', # T - A
                'R': 'Y', # R - A G - T C - Y
                'Y': 'R', # Y - C T - G A - R
                'S': 'S', # S - G C - C G - S
                'W': 'W', # W - A T - T A - W
                'K': 'M', # K - G T - C A - M
                'M': 'K', # M - A C - T G - K
                'B': 'V', # B - C G T - G C A - V
                'V': 'B', # V - A C G - T G C - B
                'D': 'H', # D - A G T - T C A - H
                'H': 'D', # H - A C T - T G A - D
                'N': 'N'
            }
            self.base_compl = {
                'A': ['T'],
                'G': ['C'],
                'C': ['G'],
                'T': ['A']
            }

    def _get_adjusted_struct(self, struct, seq):
        if struct is None:
            return '.'*len(seq)
        elif len(seq) < len(struct):
            return struct[:len(seq)]
        elif len(seq) > len(struct):
            return ''.join([struct, '.'*(len(seq) - len(struct))])
        return struct

    def _get_meta_struct(self, struct):
        struct_spec   = namedtuple(
            'struct_spec', 'struct paired_dict rev_paired_dict unpaired_set folding inversefolding')
        pairing_stack = []
        meta_struct   = struct_spec(
            struct=struct.replace('x', '.'),
            paired_dict={},
            rev_paired_dict={},
            unpaired_set=set(),
            folding={},
            inversefolding={})

        for index, nt in enumerate(struct):

            if nt == '(':
                pairing_stack.append(index)

            elif nt == ')':

                try:
                    closure = pairing_stack.pop()
                except:
                    raise ValueError(
                        ' [X] Unbalanced parentheses in structure constraint')

                meta_struct.paired_dict[closure]   = index
                meta_struct.rev_paired_dict[index] = closure

            elif nt == 'x':
                meta_struct.unpaired_set.add(index)

        if pairing_stack:
            raise ValueError(
                ' [X] Unbalanced parentheses in structure constraint')

        if meta_struct.paired_dict:
            meta_struct.folding['status'] = True
            meta_struct.inversefolding['status'] = True
        elif meta_struct.unpaired_set:
            meta_struct.folding['status'] = True
            meta_struct.inversefolding['status'] = False
        else:
            meta_struct.folding['status'] = False
            meta_struct.inversefolding['status'] = False

        return meta_struct

    def _get_meta_seq(self, seq, meta_struct):
        seq = list(seq)
        meta_seq = [None] * len(seq)

        # Normalize Meta Sequence
        for i in range(len(seq)):
            try:
                if i in meta_struct.paired_dict:
                    j = meta_struct.paired_dict[i]
                    if len(self.iupac_space[seq[j]]) < len(self.iupac_space[seq[i]]):
                        seq[i] = self.iupac_compl[seq[j]]
                    elif len(self.iupac_space[seq[i]]) < len(self.iupac_space[seq[j]]):
                        seq[j] = self.iupac_compl[seq[i]]
                meta_seq[i] = set(self.iupac_space[seq[i]])
            except:
                raise ValueError(
                    ' [X] Invalid IUPAC code at index {} in sequence constraint'.format(i))

        return tuple(meta_seq)

    def _reset_candidate_kmer_set(
        self,
        candidate,
        kmer_set,
        i):
        candidate[i] = '-'
        kmer_set[i]  = ' '

    def _clear_path(
        self,
        candidate,
        tried_set,
        kmer_set,
        i,
        k):
        j = i
        while j > k:
            self._reset_candidate_kmer_set(
                candidate,
                kmer_set,
                j)
            tried_set[j] = set()
            # tried_set[j] = set(candidate[j])
            j -= 1
        else:
            self._reset_candidate_kmer_set(
                candidate,
                kmer_set,
                j)
        i = j
        return i

    def _get_rollback_index(
        self,
        meta_seq,
        meta_struct,
        i,
        homology,
        candidate,
        tried_set):
        roll_back_index = None
        # See if any position in last homology places has potential for change
        j = i
        while (j >= i - homology + 1) and (j >= 0):
            # Case ( or . with potential for change
            if not (j in meta_struct.rev_paired_dict):
                if len(meta_seq[j]) > len(tried_set[j]):
                    roll_back_index = j
                    break
            # Case ) for RNA Parts with potential for change
            else:
                l = meta_struct.rev_paired_dict[j]
                pnt = meta_seq[j]
                pnt = pnt.intersection(
                    self.base_compl[candidate[l]])
                if len(tried_set[j]) < len(pnt):
                    roll_back_index = j
                    break
            j -= 1
        # Else go to the ( for the first ) in the last homology places
        if not roll_back_index:
            j = max(0, i - homology + 1)
            k = None
            while j <= i:
                if j in meta_struct.rev_paired_dict:
                    k = j
                    break
                j += 1
            roll_back_index = meta_struct.rev_paired_dict[k]
        return roll_back_index

    def _roll_back(
        self,
        meta_seq,
        meta_struct,
        i,
        homology,
        candidate,
        tried_set,
        kmer_set,
        rbi=None):
        # Default Case: Explicit roll back
        if not rbi is None:

            # traceback to ( if corresponding ) given
            if rbi in meta_struct.rev_paired_dict:
                j = meta_struct.rev_paired_dict[rbi]
                pnt = meta_seq[rbi]
                pnt = pnt.intersection(
                    self.base_compl[candidate[j]])
                # No potential for change
                if len(pnt) == len(tried_set[i]):
                    rbi = j

            self._clear_path(
                candidate, tried_set, kmer_set, i, k=rbi)
            return rbi

        # Case 1: i is not paired upstream
        if not i in meta_struct.rev_paired_dict:
            self._reset_candidate_kmer_set(
                candidate, kmer_set, i)

        # Case 2: i is paired upstream
        else:
            roll_back_index = self._get_rollback_index(
                meta_seq, meta_struct, i, homology, candidate, tried_set)
            i = self._clear_path(
                candidate, tried_set, kmer_set, i, k=roll_back_index)

        return i

    def _get_local_roll_back_index(
        self,
        candidate,
        i,
        local_model_fn,
        verbose):
        # Prep candidate
        candidate_str = ''.join(candidate)
        candidate_str = candidate_str[:i+1]

        # Try to evaluate the local_model_fn on candidate_str
        outcome = True
        try:
            outcome = local_model_fn(candidate_str)
            if outcome in [True, False]:
                state, index = outcome, i
            else:
                state, index = outcome
        except Exception as e:
            print(' Local Model fn. failed to evaluate partial path: {}\n'.format(
                candidate_str))
            raise e # No intelligence, halt everything!

        # State satisfactory?
        try:
            assert state in [True, False]
        except Exception as e:
            print(' Local Model fn. failed to evaluate partial path: {}'.format(
                candidate_str))
            print(' Local Model fn. returned a non-boolean evaluation: {}\n'.format(
                state))
            raise e

        # Index satisfactory?
        try:
            if state == False:
                index = int(index)
                assert 0 <= index <= i
        except Exception as e:
            print(' Local Model fn. failed to evaluate partial path: {}'.format(
                candidate_str))
            print(' Local Model fn. returned a non-integer or invalid traceback index: {}\n'.format(
                index))
            raise e

        # No conflict found!
        if state:
            return None
        # Conflict Found!
        else:
            return index

    def _get_non_coding_candidate(
        self,
        meta_seq,
        meta_struct,
        homology,
        local_model_fn,
        verbose,
        jump=False,
        start_seq=None,
        allow_internal_repeat=False):

        # Setup the data structures
        candidate = ['-'] * len(meta_seq) if not start_seq else list(start_seq)
        tried_set = [set() for _ in range(len(meta_seq))]
        kmer_set  = kmerSetArray(size=len(meta_seq))

        # Setup indexing
        i               = 0
        roll_back_count = 0

        # Main backtracking code
        while -1 < i < len(meta_seq):

            # Jumping out of iteration
            if jump and roll_back_count == homology:
                candidate = None
                break

            # Try to build a candidate
            if candidate[i] == '-':

                # Phase determination
                forward = False
                # Case )
                if i in meta_struct.rev_paired_dict:
                    j = meta_struct.rev_paired_dict[i]
                    pnt = meta_seq[i]
                    pnt = pnt.intersection(
                        self.base_compl[candidate[j]])
                    if len(pnt) > len(tried_set[i]):
                        forward = True
                # Case ( and .
                else:
                    if len(tried_set[i]) < len(meta_seq[i]):
                        forward = True

                # Forward phase - A nucleotide may be chosen
                if forward:
                    # Reset roll_back_count
                    roll_back_count = 0

                    # Case ( and .
                    if not i in meta_struct.rev_paired_dict:
                        candidate[i] = self.rng.choice(
                            sorted(meta_seq[i]-tried_set[i]))
                        tried_set[i].add(candidate[i])
                        # Case (
                        if i in meta_struct.paired_dict:
                            j = meta_struct.paired_dict[i]
                            # DNA Parts
                            if self.part_type == 'DNA':
                                pnt = self.base_compl[candidate[i]][0]
                                candidate[j] = pnt
                            # RNA Parts
                            else:
                                pnt = meta_seq[j]
                                pnt = pnt.intersection(
                                    self.base_compl[candidate[i]])
                                candidate[j] = self.rng.choice(sorted(pnt))
                            tried_set[j] = set(candidate[j])
                    # Case )
                    else:
                        j = meta_struct.rev_paired_dict[i]
                        # DNA Parts
                        if self.part_type == 'DNA':
                            pnt = self.base_compl[candidate[j]][0]
                            candidate[i] = pnt
                        # RNA Parts
                        else:
                            pnt = pnt - tried_set[i]
                            candidate[i] = self.rng.choice(sorted(pnt))
                        tried_set[i].add(candidate[i])

                # Backward phase - Nucleotide choices exhausted, so traceback
                else:
                    # Update roll_back_count
                    roll_back_count += 1

                    # Reset and roll back
                    tried_set[i] = set()
                    kmer_set[i]  = ' '
                    if i > 0:
                        # Clear stuff at current index
                        self._reset_candidate_kmer_set(
                            candidate,
                            kmer_set,
                            i)
                        # Traceback to previous index
                        # since current index done
                        i = self._roll_back(
                            meta_seq,
                            meta_struct,
                            i-1,
                            homology,
                            candidate,
                            tried_set,
                            kmer_set) + 1
                    i -= 1
                    continue

            # See if built candidate[i] is valid
            # Case ( and .
            elif not i in meta_struct.rev_paired_dict:
                # Wrong nucleotide selected
                if not candidate[i] in meta_seq[i]:
                    self._reset_candidate_kmer_set(
                        candidate=candidate,
                        kmer_set=kmer_set,
                        i=i)
                    tried_set[i] = set()
                    continue
            # Case )
            elif i in meta_struct.rev_paired_dict:
                j = meta_struct.rev_paired_dict[i]
                # Paired bases are not complementary
                # or, Wrong nucleotide selected
                if (not candidate[j] in self.base_compl[candidate[i]]) or \
                   (not candidate[i] in meta_seq[i]):
                    self._reset_candidate_kmer_set(
                        candidate=candidate,
                        kmer_set=kmer_set,
                        i=i)
                    tried_set[i] = set()
                    continue

            # See if the local model function is violated
            if local_model_fn:
                rbi = self._get_local_roll_back_index(
                    candidate, i, local_model_fn, verbose)
                # Model function violated, and
                # a traceback location was determined
                if not rbi is None:
                    roll_back_count = 0 if rbi < i else roll_back_count
                    i = self._roll_back(
                        meta_seq,
                        meta_struct,
                        i,
                        homology,
                        candidate,
                        tried_set,
                        kmer_set,
                        rbi)
                    continue

            # Are either of these mers seen previously?
            mmer_seen = False
            kmer = None

            # Handle equal internal and shared repeats
            if i >= homology-1:
                # Get the kmer/rmer
                kmer = ''.join(candidate[i-homology+1:i+1])
                rmer = utils.get_revcomp(kmer)
                mmer = min(kmer, rmer)

                # Case: kmer/rkmer is an internal
                #       repeat to current part
                if not allow_internal_repeat:
                    # Direct repeat
                    if kmer in kmer_set:
                        mmer_seen = True
                    # Inverted repeat
                    elif rmer in kmer_set:
                        mmer_seen = True
                    # Palindrome repeat
                    elif kmer == rmer:
                        mmer_seen = True

                # Case: mmer is a shared repeat with
                #       a previous part
                if not mmer_seen:
                    if mmer in self.kmer_db:
                        mmer_seen = True

                # Case: mmer is a shared repeat with
                #       background
                if not mmer_seen:
                    if not self.background is None:
                        if self.background.K == homology:
                            if mmer in self.background:
                                mmer_seen = True

            # Handle background repeats
            if not mmer_seen and \
               not self.background is None and \
               homology != self.background.K:

                # Determine background K
                K = self.background.K

                # Check is warranted
                if i >= K-1:

                    # Get the kmer/rmer
                    kmer = ''.join(candidate[i-K+1:i+1])
                    rmer = utils.get_revcomp(kmer)
                    mmer = min(kmer, rmer)

                    # Actual check
                    if mmer in self.background:
                        mmer_seen = True

            # Traceback to eliminate repeat
            if mmer_seen:
                i = self._roll_back(
                    meta_seq,
                    meta_struct,
                    i,
                    homology,
                    candidate,
                    tried_set,
                    kmer_set)
                continue

            # Everything OK .. insert kmer
            if i >= homology-1:
                kmer_set[i] = kmer

            # Roll forward
            i += 1

        # Prepare to return candidate
        del kmer_set
        if candidate is None or '-' in candidate:
            return None
        else:
            return ''.join(candidate)

    def _get_opt_pass_count(self, meta_struct, synth_opt, global_model_fn):
        opt_criteria_count = 1 # Since candidate must at least be non-repetitive
        if meta_struct.folding['status']:
            opt_criteria_count += 1
        if synth_opt:
            opt_criteria_count += 1
        if global_model_fn:
            opt_criteria_count += 1
        return opt_criteria_count

    # Diagnostic function -- Shouldn't trigger on experimental changes
    def _is_non_coding_construction_verified(self, meta_seq, meta_struct, candidate):
        i = 0
        while i < len(candidate):
            if not candidate[i] in meta_seq[i]:
                return False,1,i
            if i in meta_struct.paired_dict:
                j = meta_struct.paired_dict[i]
                if not candidate[j] in self.base_compl[candidate[i]]:
                    return False,2,i
            i += 1
        return True,0,0

    def _is_synthesis_verified(self, candidate):
        return self.synthesis.evaluate(candidate)

    def _is_structure_verified(self, meta_struct, struct_type, candidate):
        struct_satisfied = 0

        # Decide which structure criteria to fulfill
        if struct_type == 'centroid':
            candidate_structs = [self.fold.evaluate_centroid(
                candidate)]
        elif struct_type == 'both':
            candidate_structs = [self.fold.evaluate_mfe(
                candidate)]
            candidate_structs.append(self.fold.evaluate_centroid(
                candidate))
        else:
            candidate_structs = [self.fold.evaluate_mfe(
                candidate)]

        for candidate_struct in candidate_structs:
            vienna_meta_struct = self._get_meta_struct(
                candidate_struct)
            # Ensure all forbidden base indices unpaired
            for bp_closed in meta_struct.unpaired_set:
                if not vienna_meta_struct.struct[bp_closed] == '.':
                    return False
            # Ensure all paired base indices paired as desired
            for bp_open in meta_struct.paired_dict:
                if not bp_open in vienna_meta_struct.paired_dict:
                    break
                else:
                    if meta_struct.paired_dict[bp_open] != vienna_meta_struct.paired_dict[bp_open]:
                        break
            else:
                struct_satisfied += 1

        # All structure criteria satisfied
        if struct_satisfied == len(candidate_structs):
            return True
        else:
            return False

    def _get_inverse_fold_candidate(self, start_seq, meta_seq, meta_struct):
        inverse_fold_seq = ''.join(char.lower() if len(meta_seq[i]) == 1 else char for i,char in enumerate(start_seq))
        return self.fold.design(
            seq=inverse_fold_seq, struct=meta_struct.struct).upper()

    def _get_verified_non_coding_candidate(self,
        homology,
        meta_seq,
        meta_struct,
        struct_type,
        synth_opt,
        local_model_fn,
        global_model_fn,
        jump_count,
        fail_count,
        verbose,
        abortion,
        allow_internal_repeat=False):

        # Setup counts and variables
        current_jump_count = 0
        current_fail_count = 0
        struct_fail_count  = 0
        synth_fail_count   = 0
        model_fail_count   = 0
        seed_seq           = None
        verified_candidate = None
        opt_pass_count     = self._get_opt_pass_count(
            meta_struct,
            synth_opt,
            global_model_fn)


        while not verified_candidate:

            # Try to get a non-repetitive candidate
            start_seq = self._get_non_coding_candidate(
                meta_seq=meta_seq,
                meta_struct=meta_struct,
                homology=homology,
                local_model_fn=local_model_fn,
                verbose=verbose,
                jump=current_jump_count < jump_count,
                start_seq=seed_seq,
                allow_internal_repeat=allow_internal_repeat)
            candidate = start_seq
            if start_seq:
                # If structure unmatched but constraint has base pairings, go for inverse-repair strategy
                if meta_struct.inversefolding['status'] and not self._is_structure_verified(
                    meta_struct, struct_type, start_seq):
                    inverse_fold_seq = self._get_inverse_fold_candidate(
                        start_seq,
                        meta_seq,
                        meta_struct)
                    candidate = self._get_non_coding_candidate(
                        meta_seq=meta_seq,
                        meta_struct=meta_struct,
                        homology=homology,
                        local_model_fn=local_model_fn,
                        verbose=verbose,
                        jump=current_jump_count < jump_count,
                        start_seq=inverse_fold_seq,
                        allow_internal_repeat=allow_internal_repeat)

            opt_count = 0

            # If valid candidate then process
            if candidate:
                current_jump_count = 0
                opt_count += 1

                # Diagnostic block -- Shouldn't trigger on experimental changes
                construction = self._is_non_coding_construction_verified(
                    meta_seq,
                    meta_struct,
                    candidate)
                construction_state = construction[0]
                error_digest = construction[1:]
                if not construction_state:
                    current_fail_count += 1
                    raise Exception(
                        'Maker built a rogue candidate: {}\nError Digest: {}\nPlease report issue to authors.'.format(
                            candidate,
                            error_digest))
                else:
                    pass

                # Synthesis optimization
                if synth_opt:
                    if self._is_synthesis_verified(candidate):
                        opt_count += 1
                    else:
                        synth_fail_count += 1

                # Global model optimization
                if global_model_fn:

                    # Try to evaluate the global_model_fn on candidate_str
                    outcome = True
                    try:
                        outcome = global_model_fn(candidate)
                    except Exception as e:
                        print(' Global Model fn. failed to evaluate complete path: {}\n'.format(
                            candidate))
                        raise e

                    # Outcome satisfactory?
                    try:
                        assert outcome in [True, False]
                    except Exception as e:
                        print(' Global Model fn. returned a non-boolean state: {}\n'.format(
                            outcome))
                        raise e

                    # Process outcome
                    if outcome: # True
                        opt_count += 1
                    else:       # False
                        model_fail_count += 1

                # Structural optimization
                if meta_struct.folding['status']:
                    if self._is_structure_verified(
                        meta_struct,
                        struct_type,
                        candidate):
                        opt_count += 1
                    else:
                        struct_fail_count += 1

                # Did everything get optimized?
                if opt_count == opt_pass_count:
                    verified_candidate = candidate
                else:
                    current_fail_count += 1

                # Failure count exceeded, terminate
                if current_fail_count == fail_count:
                    break

            # No candidate produced
            else:
                # No jumps made yet no non-repetitive candidate found
                if current_jump_count >= jump_count:
                    break
                # Increase current_jump_count
                else:
                    current_jump_count += 1
                # Abortion limit reached?
                if abortion and current_jump_count >= jump_count:
                    break

        if int(verbose) > 1:
            print('\n  [seq fails] {}, [struct fails] {}, [synth fails] {}, [global fails] {}, [opt fails] {}'.format(
                current_jump_count,
                struct_fail_count,
                synth_fail_count,
                model_fail_count,
                current_fail_count))

        return verified_candidate, current_jump_count+1, current_fail_count+1

    def _get_non_coding_nrps(
        self,
        homology,
        seq,
        struct,
        struct_type,
        target,
        synth_opt,
        local_model_fn,
        global_model_fn,
        jump_count,
        fail_count,
        verbose,
        abortion,
        allow_internal_repeat=False):

        # Setup structures
        seq = seq.upper()
        meta_struct = self._get_meta_struct(
            struct=self._get_adjusted_struct(struct, seq))
        meta_seq = self._get_meta_seq(
            seq=seq,
            meta_struct=meta_struct)

        seq_count  = 0
        iter_count = 0
        time_sum   = 0.0
        begin_time = time()
        break_flag = False

        # Setup Bernoulli Success model
        total_jump_trials    = jump_count
        total_jump_successes = 1
        curr_jump_prob       = berno.get_prob(
            trials=total_jump_trials,
            success=total_jump_successes)
        curr_jump_trial      = jump_count
        total_fail_trials    = fail_count
        total_fail_successes = 1
        curr_fail_prob       = berno.get_prob(
            trials=total_fail_trials,
            success=total_fail_successes)
        curr_fail_trial      = fail_count

        # Stream parts until completion
        while True:

            t0 = time()

            candidate, curr_jump_trial, curr_fail_trial = self._get_verified_non_coding_candidate(
                homology,
                meta_seq,
                meta_struct,
                struct_type,
                synth_opt,
                local_model_fn,
                global_model_fn,
                curr_jump_trial,
                curr_fail_trial,
                verbose,
                abortion,
                allow_internal_repeat)

            if candidate is None:
                break_flag = True

            # Got a canidate -- will try again
            if not break_flag:
                final_candidate = candidate
                update_status   = True
                try:
                    for kmer in utils.stream_min_kmers(
                        seq=candidate,
                        k=homology):
                        self.kmer_db.add(kmer)
                except Exception as E:
                    update_status = False

                if not update_status: # Memory full
                    if verbose:
                        print('[ERROR] Memory Full ... Breaking Loop')
                        yield update_status

                seq_count  += 1
                time_sum   += time()-t0
                iter_count += 1

                if verbose:
                    print(' [part] {}, [{}-mers] {}, [iter time] {:.2f}s, [avg time] {:.2f}s, [total time] {:.2f}h'.format(
                        seq_count,
                        homology,
                        len(self.kmer_db),
                        time()-t0, time_sum / iter_count,
                        (time() - begin_time) / 3600.0))

                yield final_candidate

                # No more parts required
                if seq_count == target:
                    break
            # No more candidates to build
            else:
                yield candidate
                break

            # Update failure limits based on Bernoulli Success model
            total_jump_trials    += curr_jump_trial
            total_jump_successes += 1
            curr_jump_prob       =  berno.get_prob(trials=total_jump_trials, success=total_jump_successes)
            curr_jump_trial      =  berno.get_trials(prob=curr_jump_prob)
            total_fail_trials    += curr_fail_trial
            total_fail_successes += 1
            curr_fail_prob       =  berno.get_prob(trials=total_fail_trials, success=total_fail_successes)
            curr_fail_trial      =  berno.get_trials(prob=curr_fail_prob)

    def _check_maker_constraints(
        self,
        seq,
        struct,
        part_type,
        allow_internal_repeat,
        target,
        homology):
        # Sequence Legality 1
        if not isinstance(seq, str):
            print('\n [ERROR]    Sequence Constraint must be a string, not {}'.format(type(seq)))
            print(' [SOLUTION] Try correcting Sequence Constraint\n')
            return False
        # Sequence Legality 2
        if len(seq) < 5:
            print('\n [ERROR]    Sequence Constraint must be longer than 4 bases, not {}'.format(len(seq)))
            print(' [SOLUTION] Try using a longer Sequence Constraint\n')
            return False
        # Structure Legality 1
        if not isinstance(struct, str):
            print('\n [ERROR]    Structure Constraint must be a string, not {}'.format(type(struct)))
            print(' [SOLUTION] Try correcting Structure Constraint\n')
            return False
        # Structure Legality 2
        if len(struct) != len(seq):
            print('\n [ERROR]    Structure Constraint must be same length as Sequence Constraint ({}), not {}'.format(
                len(seq),
                len(struct)))
            print(' [SOLUTION] Try correcting length of Structure Constraint\n')
            return False
        # Part Type Legality 1
        if not isinstance(part_type, str):
            print('\n [ERROR]    Part Type must be a string, not \'{}\''.format(type(part_type)))
            print(' [SOLUTION] Try correcting Part Type\n')
            return False
        # Part Type Legality 2
        if not part_type in ['DNA', 'RNA']:
            print('\n [ERROR]    Part Type must be \'RNA\' or \'DNA\', not \'{}\''.format(part_type))
            print(' [SOLUTION] Try correcting Part Type\n')
            return False
        # Sequence Legality 3
        seq_legal, seq_illegal_chars = makerchecks.is_seq_constr_legal(seq, part_type)
        if not seq_legal:
            print('\n [ERROR]    {} Sequence Constraint is not legal due to chars: {}'.format(part_type, seq_illegal_chars))
            print(' [SOLUTION] Try correcting Sequence Constraint or Part Type\n')
            return False
        # Structure Legality 3
        struct_legal, unclosed, unopened, invalid = makerchecks.is_structure_valid(struct)
        if not struct_legal:
            print('\n [ERROR]    Structure Constraint is illegal or unbalanced')
            if unclosed:
                print(' [ERROR]    >> Unclosed bases at locations: {}'.format(unclosed))
            if unopened:
                print(' [ERROR]    >> Unopened bases at locations: {}'.format(unopened))
            if invalid:
                print(' [ERROR]    >> Invalid characters at locations: {}'.format(invalid))
            print(' [SOLUTION] Try correcting Structure Constraint\n')
            return False
        # Sequence + Structure + Part Type Base Pairing Combination Legality
        combo_state, incompat_locs, reduced_locs = makerchecks.is_pairing_compatible(seq, struct, part_type)
        if not combo_state:
            if incompat_locs:
                print('\n [ERROR]    Incompatible Base Pairing for {} Parts at locations: {}'.format(part_type, incompat_locs))
                print(' [SOLUTION] Try correcting Sequence Constraint or Part Type\n')
                return False
            if reduced_locs: # -- soft check
                print('\n [WARNING]  Reducible Paired Bases at locations: {}'.format(reduced_locs))
                print(' [WARNING]  Fewer Parts may be Generated')
        # Lmax Legality 1
        if not isinstance(homology, int):
            print('\n [ERROR]    Lmax must be an integer, not {}'.format(type(homology)))
            print(' [SOLUTION] Try correcting Lmax\n')
            return False
        # Lmax Legality 2
        if homology-1 < 5:
            print('\n [ERROR]    Lmax must be greater than 4, not {}'.format(homology-1))
            print(' [SOLUTION] Try correcting Lmax\n')
            return False
        # Lmax Legality 3
        if homology-1 >= len(seq):
            print('\n [ERROR]    Lmax must be less than length of Sequence Constraint ({}), not {}'.format(len(seq), homology-1))
            print(' [SOLUTION] Try correcting Lmax\n')
            return False
        # Target Size Legality 1
        if not isinstance(target, int):
            print('\n [ERROR]    Target Size must be an integer, not {}'.format(type(target)))
            print(' [SOLUTION] Try correcting Target Size\n')
            return False
        # Target Size Legality 2
        if target < 1:
            print('\n [ERROR]    Target Size must be greater than 0, not {}'.format(target))
            print(' [SOLUTION] Try correcting Target Size\n')
            return False
        # Sequence Sufficiency -- soft check
        seq_sufficient, constrained_motif_locs = makerchecks.is_seq_constr_sufficient(seq, struct, homology, target)
        if not seq_sufficient:
            print('\n [WARNING]  Target Size of {} may be unreachable from given Sequence/Structure Constraint and Lmax of {}'.format(target, homology-1))
            print(' [WARNING]  >> Lmax limiting windows between locations: {}'.format(constrained_motif_locs))
            print(' [WARNING]  Fewer Parts may be Generated')
        # Internal Repeats Legality
        if not allow_internal_repeat in [True, False]:
            print('\n [ERROR]    Internal Repeat must be boolean, not {}'.format(type(allow_internal_repeat)))
            print(' [SOLUTION] Try correcting Internal Repeat\n')
            return False
        # Structure Sufficiency
        if not allow_internal_repeat:
            struct_sufficient, long_hairpins = makerchecks.is_structure_not_conflict(struct, homology)
            if not struct_sufficient:
                print('\n [ERROR] Structure Constraint is insufficient based on given Lmax')
                for long_hairpin in long_hairpins:
                    print(' [ERROR] >> Long hairpin at locations: {}'.format(long_hairpin))
                    print(' [SOLUTION] Try relaxing Structure Constraint or setting internal_repeats=True')
                print()
                return False
        return True

    def _check_maker_inputs(
        self,
        struct_type,
        synth_opt,
        jump_count,
        fail_count,
        output_file,
        verbose):
        # Struct Type Legality
        if not struct_type in ['mfe', 'centroid', 'both']:
            print('\n [ERROR]    Struct Type must be \'mfe\', \'centroid\' or \'both\', not \'{}\''.format(struct_type))
            print(' [SOLUTION] Try correcting Part Type\n')
            return False
        # Synth Optimization Legality
        if not synth_opt in [True, False]:
            print('\n [ERROR]    Synth Opt must be True or False, not {}'.format(synth_opt))
            print(' [SOLUTION] Try correcting Synth Opt\n')
            return False
        # Jump Count Legality 1
        if not isinstance(jump_count, int):
            print('\n [ERROR]    Jump Count must be an integer, not {}'.format(type(jump_count)))
            print(' [SOLUTION] Try correcting Jump Count\n')
            return False
        # Jump Count Legality 2
        if jump_count < 0:
            print('\n [ERROR]    Jump Count must be greater than 0, not {}'.format(jump_count))
            print(' [SOLUTION] Try correcting Jump Count\n')
            return False
        # Fail Count Legality 1
        if not isinstance(fail_count, int):
            print('\n [ERROR]    Fail Count must be an integer, not {}'.format(type(fail_count)))
            print(' [SOLUTION] Try correcting Fail Count\n')
            return False
        # Fail Count Legality 2
        if fail_count < 0:
            print('\n [ERROR]    Fail Count must be greater than 0, not {}'.format(fail_count))
            print(' [SOLUTION] Try correcting Fail Count\n')
            return False
        # Output File Legality
        if not output_file is None:
            if not isinstance(output_file, str):
                print('\n [ERROR]    Output File must be a string or None, not {}'.format(type(output_file)))
                print(' [SOLUTION] Try correcting Output File\n')
                return False
        # Everything OK
        return True

    def nrp_maker(self,
        homology,
        seq_constr,
        struct_constr,
        target_size,
        background=None,
        struct_type=None,
        synth_opt=True,
        local_model_fn=None,
        global_model_fn=None,
        jump_count=10,
        fail_count=1000,
        output_file=None,
        verbose=True,
        abortion=True,
        allow_internal_repeat=False,
        check_constraints=True):

        if verbose:
            print('\n[Non-Repetitive Parts Calculator - Maker Mode]')
        build_parts = True

        # Check Maker Constraints
        if check_constraints:
            if verbose:
                print('\n[Checking Constraints]')
                print('  Sequence Constraint: {}'.format(seq_constr))
                print(' Structure Constraint: {}'.format(struct_constr))
                print('      Part Type      : {}'.format(self.part_type))
                print('           Lmax      : {} bp'.format(homology-1))
                print('    Target Size      : {} parts'.format(target_size))
                print('  Internal Repeats   : {}'.format(allow_internal_repeat))
            check_status = self._check_maker_constraints(
                seq_constr,
                struct_constr,
                self.part_type,
                allow_internal_repeat,
                target_size,
                homology)

            if check_status == False:
                if verbose:
                    print(' Check Status: FAIL\n')
                build_parts = False
            else:
                if verbose:
                    print('\n Check Status: PASS')

            # Background Check
            if build_parts:
                if not background is None:
                    if verbose:
                        print('\n[Checking Background]\n Background: {}'.format(background))
                    if isinstance(background, kmerSetDB):
                        if background.K > len(seq_constr):
                            build_parts = False
                            print('\n [ERROR]    Background Lmax of {} is greater than desired part length ({}-bp)'.format(
                                background.K-1,
                                len(seq_constr)))
                            print(' [SOLUTION] Try using a background with Lmax less than or equal to part length\n')
                            if verbose:
                                print(' Check Status: FAIL\n')
                        if build_parts and not background.ALIVE:
                            build_parts = False
                            print('\n [ERROR]    Background is closed or dropped')
                            print(' [SOLUTION] Try using an open Background\n')
                            if verbose:
                                print(' Check Status: FAIL\n')
                        if build_parts:
                            if verbose:
                                print('\n Check Status: PASS')

                    else:
                        build_parts = False
                        print('\n [ERROR]    Background Object is INVALID')
                        print(' [SOLUTION] Try instantiating background via nrpcalc.background(...)\n')
                        if verbose:
                            print(' Check Status : FAIL\n')

            # Arguments Check
            if build_parts:
                if verbose:
                    print('\n[Checking Arguments]')
                    print(' Struct Type : {}'.format(struct_type))
                    print('  Synth Opt  : {}'.format(synth_opt))
                    print('   Jump Count: {}'.format(jump_count))
                    print('   Fail Count: {}'.format(fail_count))
                    print(' Output File : {}'.format(output_file))
                check_status = self._check_maker_inputs(
                    struct_type,
                    synth_opt,
                    jump_count,
                    fail_count,
                    output_file,
                    verbose)

                if check_status == False:
                    if verbose:
                        print(' Check Status: FAIL\n')
                    build_parts = False
                else:
                    if verbose:
                        print('\n Check Status: PASS')

            if not build_parts:
                # Cleanups
                self.background = None
                self.kmer_db = None
                raise RuntimeError('Invalid Constraints, Background or Arguments')

        # Separate Checks from Build Logs
        if verbose:
            print()

        # kmer_db and background Setup
        projector.setup_proj_dir(self.proj_id)
        self.kmer_db = set()
        self.background = background
        # self.kmer_db = kmerSetDB(
        #     path='./{}/kmerDB'.format(self.proj_id),
        #     homology=homology,
        #     verbose=verbose)

        # Project Setup
        if output_file is None:
            projector.setup_proj_dir(self.proj_id)
            output_file = './{}/seq_list.fa'.format(self.proj_id)

        # Execute Maker
        with open(output_file, 'w') as out_file:
            seq_constr    = seq_constr.upper()
            struct_constr = self._get_adjusted_struct(
                struct_constr,
                seq_constr)
            current_nrp_count = 0
            memory_exhausted  = False

            if verbose:
                print('Constructing Toolbox:\n')

            for non_coding_nrp in self._get_non_coding_nrps(
                homology,
                seq_constr,
                struct_constr,
                struct_type,
                target_size,
                synth_opt,
                local_model_fn,
                global_model_fn,
                jump_count,
                fail_count,
                verbose,
                abortion,
                allow_internal_repeat):

                # Write out genetic part
                if non_coding_nrp:
                    out_file.write(
                        '>non-repetitive part {}\n'.format(
                            current_nrp_count+1))
                    non_coding_nrp = '\n'.join(
                        textwrap.wrap(
                            non_coding_nrp, 80))
                    out_file.write('{}\n'.format(
                        non_coding_nrp))
                    current_nrp_count += 1
                else:
                    if non_coding_nrp is None:
                        if verbose:
                            print('Failure Limits Exceeded or k-mers Exhausted. Cannot Build More Parts.')
                    else:
                        if verbose:
                            print('Memory Capacity at Full. Cannot Build More Parts.')
                        memory_exhausted = True

                # Memory no longer available ... stop
                if memory_exhausted:
                    break
            if verbose:
                print('\nConstruction Complete.\n')

        # Detach Background
        self.background = None

        # Remove kmerSetDB
        # self.kmer_db.drop()
        self.kmer_db = None

        # Pack output in dictionary
        parts_dict = {}
        for i,line in enumerate(utils.stream_fasta_seq_list(output_file)):
            line = line.strip()
            parts_dict[i] = line

        # Cleanups and Return
        projector.remove_proj_dir()
        if verbose:
            print('Non-Repetitive Toolbox Size: {}'.format(current_nrp_count))
        return parts_dict

def main():
    homology = 16
    sm_obj = NRPMaker(seed=7)

    # Hammerhead Nielsen Paper
    seq         = 'NNNN AGNNNU CANNNNN UGUGCUU NNNNNU CUGAUGA NNNN GUGA NNNN GAAA NNNC CUCU NNNNN UAAU NNNNN UUAA NNNN' # Nielsen Like
    struct      = 'xxxx x((((( x(((((( xxxxxxx )))))) xxxxxxx (((( xxxx )))) xxx) )))) xxxx ((((( xxxx ))))) xxxx xxxx'
    seq         = ''.join(seq.split(' '))
    struct      = ''.join(struct.split(' '))
    output_file = None #'riboz.fa'
    background  = None #utils.get_fasta_seq_list(fasta_filename='input.fa.bk104')

    # Final Result Store
    final_toolbox = {}

    # Initialize Background
    background = kmerSetDB(
        path='./testDB',
        homology=homology,
        verbose=True)
    # background.multiadd(
    #     utils.get_fasta_seq_list(
    #         fasta_filename='input.fa.bk104'))

    # Background Based Single Part Design Works
    t0 = time()
    tt = 0
    toolbox1 = sm_obj.nrp_maker(homology, [seq], [struct], [10],
        struct_type='mfe',
        background=background,
        jump_count=100,
        fail_count=1000,
        synth_opt=False,
        verbose=True,
        abortion=True,
        allow_internal_repeat=True,
        output_file=output_file)
    tt += time() - t0
    final_toolbox.update(
        zip(range(len(final_toolbox), len(final_toolbox)+len(toolbox1)),
            toolbox1.values()))

    # Adding More Background Post Design Works
    background.multiadd(toolbox1.values())

    # Serial Designs from Multiple Constraints Works
    t0 = time()
    toolbox2 = sm_obj.nrp_maker(homology, [seq]*2, [struct]*2, [20]*2,
        struct_type='mfe',
        background=background,
        jump_count=100,
        fail_count=1000,
        synth_opt=False,
        verbose=True,
        abortion=True,
        allow_internal_repeat=False,
        output_file=output_file)
    tt += time() - t0
    final_toolbox.update(
        zip(range(len(final_toolbox), len(final_toolbox)+len(toolbox2)),
            toolbox2.values()))

    # Assert All Parts Unique and Non-Repetitive
    assert len(set(final_toolbox.values())) == len(
        finder.nrp_finder(final_toolbox.values(), homology, None, verbose=False))

    # Drop Background
    background.drop()

    # Report Time Elapsed
    print('\nWall Time {} sec'.format(tt))

if __name__ == '__main__':
    main()



import os
import sys
from typing import Tuple

import pkg_resources

from Bio import SeqIO

import RNA

import numpy as np

complement_table = str.maketrans('ATGCU', 'TACGA')

def stream_fasta_seq_list(fasta_filename):
    with open(fasta_filename, "rU") as handle:
        for record in SeqIO.parse(handle, "fasta"):
            yield str(record.seq)

def get_fasta_seq_list(fasta_filename):
    return list(stream_fasta_seq_list(fasta_filename))

def stream_txt_seq_list(text_filename):
    with open(text_filename) as infile:
        for line in infile:
            yield line.strip()

def get_txt_seq_list(text_filename):
    return list(stream_txt_seq_list(text_filename))

def uniquify_background_list(background_list):
    uniq_background_set = set()
    while background_list:
        uniq_background_set.add(background_list.pop())
    background_list = []
    while uniq_background_set:
        background_list.append(uniq_background_set.pop())
    return background_list

def stream_kmers(seq, k):
    if k >= len(seq):
        return [seq]
    return (seq[i:i+k] for i in range(len(seq)-k+1))

def get_comp(seq):
    return seq.translate(complement_table)

def get_revcomp(seq):
    return get_comp(seq)[::-1]

def stream_min_kmers(seq, k):
    for kmer in stream_kmers(seq, k):
        yield min(kmer, get_revcomp(kmer))

class Fold(object):

    def __init__(
        self,
        temp=37.0,
        dangles=2,
        part_type='RNA'):

        if not part_type in ['RNA', 'DNA']:
            part_type = 'RNA'

        if part_type == 'DNA':
            RNA.cvar.noGU        = True
            RNA.cvar.noGUclosure = True

        self.parameter_directory = os.path.dirname(
            os.path.abspath(__file__))#"/usr/local/share/ViennaRNA/"
        # Temperature in Celsius;
        # default=37.0 (float)
        RNA.cvar.temperature = temp
        # Dangling end energies (0,1,2);
        # see RNAlib documentation;
        # default=2 (int)
        RNA.cvar.dangles = dangles
        self.settings    = RNA.md()

        self.part_type = part_type

        parameter_file = pkg_resources.resource_filename(
            'nrpcalc', 'base/{}.par'.format(
                self.part_type))
        RNA.read_parameter_file(parameter_file)

        if part_type == 'DNA':
            self.clear_warning()

        self.adjust = self.adjust_dG(temp)

    def adjust_dG(self, temp):
        # Adjustment according to Dirks et al.

        kB = 0.00198717 # Boltzmann constant in kcal/mol/K
        T = temp
        a = [-3.983035, 301.797, 522528.9, 69.34881, 999.974950]

        # Calculate the number of moles of water per liter (molarity) at temperature (T in deg C)
        # Density of water calculated using data from
        # Tanaka M., Girard, G., Davis, R., Peuto A., Bignell, N.
        # Recommended table for the density of water..., Metrologia, 2001, 38, 301-309
        pH2O = a[4] * (
            1 - (T+a[0])**2 * (T+a[1]) / \
                (a[2]) /                 \
                (T+a[3])) /              \
            18.0152

        return -kB * (T + 273.15) * np.log(pH2O)

    def clear_warning(self):
        clrlen = len('WARNING: stacking enthalpies not symmetric')
        sys.stdout.write('\033[F\033[F\033[F\033[F')
        sys.stdout.write(' '*clrlen+'\n')
        sys.stdout.write(' '*clrlen+'\n')
        sys.stdout.write(' '*clrlen+'\n')
        sys.stdout.write(' '*clrlen+'\n')
        sys.stdout.write('\033[F\033[F\033[F\033[F')
        sys.stdout.flush()

    def evaluate_mfe(self, seq, dg=False):
        # MFE Structure Only
        fc_obj = RNA.fold_compound(seq, self.settings)
        struct,energy = fc_obj.mfe()
        if not dg:
            return struct
        else:
            return struct, energy

    def evaluate_centroid(self, seq, dg=False):
        # Centroid Structure Only
        fc_obj = RNA.fold_compound(seq, self.settings)
        fc_obj.pf()
        struct,energy = fc_obj.centroid()
        if not dg:
            return struct
        else:
            return struct, energy

    def design(self, seq, struct):
        # Closest MFE Structure Sequence
        inv = RNA.inverse_fold(seq, struct)[0]
        if self.part_type == 'DNA':
            inv = inv.replace('U', 'T').replace('u', 't')
        return inv

    def evaluate_mfe_dimer(self, seq1, seq2):
        # MFE Dimer Structure and Energy
        fc_obj = RNA.fold_compound(seq1+'&'+seq2, self.settings)
        struct,energy = fc_obj.mfe_dimer()
        struct1 = struct[:len(seq1)]
        struct2 = struct[len(seq1):]
        energy += self.adjust
        return (struct1, struct2, energy)

if __name__ == '__main__':
    pass


import sys

from . import utils

import networkx as nx

from time             import time
from itertools        import islice
from collections      import deque, defaultdict

def uniquify_seq_list(seq_list):
    unique_seq_index = {}
    index = len(seq_list)-1
    while index > -1:
        seq = seq_list.pop().upper()
        unique_seq_index[seq] = index
        index -= 1
    unique_seq_list = []
    while unique_seq_index:
        unique_seq_list.append(unique_seq_index.popitem()[::-1])
    return unique_seq_list

def get_repeat_dict(
    seq_deque,
    background,
    homology,
    internal_repeats,
    verbose):

    repeat_dict  = {}

    if verbose:
        clear_length = 0

    while seq_deque:

        if verbose:
            printed = ' [Sequence processing remaining] = {}\r'.format(
                len(seq_deque))
            sys.stdout.write(' '*clear_length+'\r')
            sys.stdout.write(printed)
            clear_length = len(printed)
            sys.stdout.flush()

        seq_id, seq = seq_deque.popleft()
        kmers = list(utils.stream_kmers(seq, k=homology))
        rmers = list(map(utils.get_revcomp, kmers))

        # Resolve internal repeats
        if not internal_repeats:
            kmer_set = set(kmers)

            # Resolve direct internal repeat
            if len(kmer_set) < len(kmers):
                # print('----->>>>')
                # print(seq, 'direct')
                continue

            inverted_repeats_found = False
            for rmer in rmers:
                if rmer in kmer_set:
                    inverted_repeats_found = True
                    break

            # Resolve inverted internal repeat
            if inverted_repeats_found:
                # print('----->>>>')
                # print(seq, 'inverted')
                continue

        # Resolve background repeats
        if not background is None:
            if background.K <= len(seq):
                if background.K == homology:
                    if any(background.multicheck(kmers)):
                        continue
                else:
                    if seq in background:
                        continue

        # Populate repeat_dict for sequences sharing repeats
        for i in range(len(kmers)):
            hmer = kmers[i]
            rmer = rmers[i]
            mmer = min(hmer, rmer)
            if mmer in repeat_dict:
                repeat_dict[mmer].append(seq_id)
            else:
                repeat_dict[mmer] = [seq_id]

    if verbose:
        print()

    return repeat_dict

def get_maximal_cliques(repeat_cliques_list):
    # Note: https://goo.gl/2NMKzh
    element_clique_index = defaultdict(set)
    for i,clique in enumerate(repeat_cliques_list):
        for element in clique:
            element_clique_index[element].add(i)
    maximal_cliques_set = set()
    for i, clique in enumerate(repeat_cliques_list):
        maximal_clique = set()
        for element in clique:
            if len(maximal_clique) == 0:
                maximal_clique |= element_clique_index[element]
            else:
                maximal_clique &= element_clique_index[element]
        if len(maximal_clique) == 1: # Current set has no superset
            maximal_cliques_set.add(i)
        else:                        # Current set has a superset
            maximal_cliques_set.update(
                maximal_clique - set([i]))
    return [repeat_cliques_list[i] for i in maximal_cliques_set]

def get_repeat_cliques(
    seq_deque,
    background,
    homology,
    internal_repeats,
    verbose):
    repeat_dict = get_repeat_dict(
        seq_deque,
        background,
        homology,
        internal_repeats,
        verbose)

    # Get unique cliques
    repeat_cliques_set = set()
    while repeat_dict:
        hmer, repeat_clique = repeat_dict.popitem()
        # Note: each repeat_clique is sorted by construction
        repeat_cliques_set.add(tuple(repeat_clique))

    # Transform clique tuples into sets
    repeat_cliques_list = list()
    while repeat_cliques_set:
        repeat_cliques_list.append(
            set(repeat_cliques_set.pop()))

    # Get maximal cliques
    repeat_cliques_list = get_maximal_cliques(
        repeat_cliques_list)

    # Compute new repeat cliques set
    repeat_cliques_set = set()
    while repeat_cliques_list:
        repeat_cliques_set.add(
            tuple(repeat_cliques_list.pop()))

    # Stream maximal repeat cliques
    while repeat_cliques_set:
        yield deque(repeat_cliques_set.pop())

def build_homology_graph(cliques, verbose):
    homology_graph = nx.Graph()

    if verbose:
        print()
        clear_length = 0

    i = 0
    while True:
        # Stream Clique
        try:
            repeat_clique = next(cliques)
        except Exception as E:
            break

        # Insert Clique
        if len(repeat_clique) == 1:
            homology_graph.add_node(
                repeat_clique.popleft())
        else:
            repeat_clique_set = set(repeat_clique)
            while repeat_clique:
                vi = repeat_clique.popleft()
                repeat_clique_set.remove(vi)
                homology_graph.add_node(vi)
                new_neighbors = repeat_clique_set - set(homology_graph[vi])
                for vj in new_neighbors:
                    homology_graph.add_edge(vi, vj)

        # Verbose Output
        if verbose:
            printed = ' [Cliques inserted] = {}\r'.format(i+1)
            sys.stdout.write(' '*clear_length+'\r')
            sys.stdout.write(printed)
            clear_length = len(printed)
            sys.stdout.flush()

        # Update Insertion Count
        i += 1

    if verbose:
        print()

    return homology_graph

def get_homology_graph(
    seq_list,
    background,
    homology,
    internal_repeats,
    seq_file,
    verbose):
    t0 = time()
    num_seqs = len(seq_list)
    seq_list = uniquify_seq_list(seq_list)
    unq_seqs = len(seq_list)
    if verbose:
        print('Extracted {} unique sequences out of {} sequences in {:2.4} seconds'.format(
            unq_seqs, num_seqs, time()-t0))

    t0 = time()
    with open(seq_file, 'w') as outfile:
        seq_deque = deque()
        while seq_list:
            seq_id,seq = seq_list.pop()
            seq_deque.append((seq_id, seq))
            outfile.write('{},{}\n'.format(seq_id, seq))
    if verbose:
        print('\nWritten {} unique sequences out to {} in {:2.4} seconds'.format(
            unq_seqs, seq_file, time()-t0))

    t0 = time()
    repeat_cliques = get_repeat_cliques(
        seq_deque,
        background,
        homology,
        internal_repeats,
        verbose)

    homology_graph = build_homology_graph(
        repeat_cliques,
        verbose)

    if verbose:
        print('\nBuilt homology graph in {:2.4} seconds. [Edges = {}] [Nodes = {}]'.format(
            time()-t0,
            homology_graph.number_of_edges(),
            homology_graph.number_of_nodes()))
        print(' [Intital Nodes = {}] - [Repetitive Nodes = {}] = [Final Nodes = {}]'.format(
            unq_seqs,
            unq_seqs - homology_graph.number_of_nodes(),
            homology_graph.number_of_nodes()))
    return homology_graph


from math          import log, ceil
from scipy.special import betaincinv as ribf

import random

zero_const = -24 # Lower is Rarer # Originally: 1. / 10**10

def get_trials(prob):
    return 2 * int(ceil(-24 / log(1. - prob)))

def get_prob(trials, success):
    # probability for rare event success
    return (success + 1.) / (trials + 2.)

def get_CI(trials, success, conf_interval):
    x1 = ribf(success + 1, trials - success + 1, 0.5*(1 - conf_interval))
    x2 = ribf(success + 1, trials - success + 1, 0.5*(1 + conf_interval))
    return (x1, x2)

def test_func():
    if random.random() <= 0.002: #1. / 200:
        return True
    return False

def main():
    curr_prob     = 1. / 100000
    conf_interval = 0.9999

    total_success = 1.
    total_trials  = get_init_trials(recommended_trials=50)
    curr_prob     = get_prob(total_trials, total_success)
    new_trial     = total_trials

    attempt       = 0

    while True:

        for t in xrange(new_trial):
            total_trials += 1
            if test_func():
                total_success += 1.
                print('[{}] Current Prob = {}'.format(attempt, get_prob(total_trials, total_success)))
                attempt += 1
                break
        else:
            print()
            print('  Est Prob = {}'.format(get_prob(total_trials, total_success)))
            x1, x2 = get_CI(total_trials, total_success, conf_interval)
            print('-99.99% CI = {}'.format(x1))
            print('+99.99% CI = {}'.format(x2))
            return None

        curr_prob = get_prob(total_trials, total_success)
        new_trial = get_trials(curr_prob)

if __name__ == '__main__':
    main()


from setuptools import setup
from os import path
from io import open

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name='nrpcalc',

    # Link: https://www.python.org/dev/peps/pep-0440/#version-scheme
    version='1.6.3',

    description='Non-Repetitive Parts Calculator - Automated design and discovery of non-repetitive genetic parts for engineering stable genetic systems',

    long_description=long_description,

    long_description_content_type='text/markdown',

    url='https://github.com/ayaanhossain/nrpcalc',

    author='Ayaan Hossain and Howard Salis',

    author_email='auh57@psu.edu, salis@psu.edu',  # Optional

    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Chemistry',

        # Pick your license as you wish
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        # These classifiers are *not* checked by 'pip install'. See instead
        # 'python_requires' below.
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    keywords=' '.join([
        'synthetic',
        'computational',
        'biology',
        'genetic',
        'parts',
        'calculator',
        'non-repetitive',
        'design',
        'discovery',
        'algorithm',
        'stable',
        'systems',
        'nrp',
        'repeats',
        'vertex',
        'cover',
        'path',
        'finding']),

    packages=['nrpcalc', 'nrpcalc.base'],

    package_dir={
        'nrpcalc': './nrpcalc'
    },

    package_data={
        'nrpcalc': ['base/*.par']
    },

    python_requires=', '.join([
        '!=2.7',
        '!=3.0.*',
        '!=3.1.*',
        '!=3.2.*',
        '!=3.3.*',
        '!=3.4.*',
        '!=3.5.*',
        '>=3.6.*',
        '<4.0.*']),

    install_requires=[
        'numpy>=1.19.0',
        'biopython>=1.77',
        'leveldb>=0.201',
        'scipy>=1.5.1',
        'networkx>=2.4',
        'jupyter>=1.0.0',
        'scikit-learn>=0.23.1',
        'seaborn>=0.10.1',
        'statsmodels>=0.11.0',
        'ViennaRNA>=2.4.18a1'],

    project_urls={  # Optional
        'Bug Reports': 'https://github.com/ayaanhossain/nrpcalc/issues',
        'Source'     : 'https://github.com/ayaanhossain/nrpcalc/tree/master/nrpcalc',
    },
)


from .nrpcalc import __authors__
from .nrpcalc import __version__
from .nrpcalc import __doc__
from .nrpcalc import background
from .nrpcalc import finder
from .nrpcalc import maker

__authors__ = __authors__
__version__ = __version__
__doc__     = __doc__



import sys

from . import vercov

import networkx as nx

from itertools import count
from time      import time
from math      import log10

def is_graph_empty(homology_graph):
    if not homology_graph.number_of_nodes():
        return True
    return False

def get_vercov_func(vercov_func, homology_graph):
    # TO DO: NEED TO USE DIFFERENT FUNCTIONS AT DIFFERENT EDGE COUNT SCALES!!
    if vercov_func == 'nrpG':
        return vercov.nrp_vercov_greedy, 'NRP Greedy'
    elif vercov_func == '2apx':
        # return vercov.std_vercov_approx, 'Standard 2-approximation'
        return vercov.nx_vercov, 'Standard 2-approximation'
    else:
        return vercov.nrp_vercov_approx, 'NRP 2-approximation'

def dump_homology_graph(homology_graph, graph_file):
    nx.write_adjlist(homology_graph, graph_file)

def load_homology_graph(graph_file):
    return nx.read_adjlist(graph_file, nodetype=int)

def get_recovered_non_homologs(homology_graph, graph_file, vercov_func=None, verbose=True):
    indiset_nodes   = set()
    # powertex_elimination(homology_graph, verbose)
    completex_nodes = [] #completex_elimination(homology_graph, verbose)
    # indiset_nodes.update(completex_nodes)

    possible_nodes  = set(homology_graph.nodes())

    if verbose:
        print('\n [+] Initial independent set = {}, computing vertex cover on remaining {} nodes.'.format(len(indiset_nodes), len(completex_nodes), len(possible_nodes)))

    if is_graph_empty(homology_graph):
        if verbose:
            print(' [X] Graph is empty, further independent set expansion not possible, terminating.')
    else:
        vercov_func, vercov_func_name = get_vercov_func(vercov_func, homology_graph)
        iteration = -1

        if verbose:
            print(' [+] Vertex Cover Function: {}'.format(vercov_func_name))
            sys.stdout.write(' [+] Dumping graph into: {}'.format(graph_file))

        t0 = time()
        dump_homology_graph(homology_graph, graph_file)
        if verbose:
            sys.stdout.write(' in {} seconds\n'.format(time()-t0))

        while True:
            iteration += 1

            if verbose:
                print('\n----------------------')
                print('Now running iteration: {}'.format(iteration))
                print('----------------------')

            t0 = time()

            if iteration > 0:
                homology_graph = nx.Graph(homology_graph.subgraph(possible_nodes))

            vercov_nodes = vercov_func(homology_graph, verbose)

            if verbose:
                print('\n [+] Computed vertex cover of size: {} (in {:.4} seconds)'.format(len(vercov_nodes), time()-t0))
                print(' [+] Loading graph from: {}'.format(graph_file))

            homology_graph = load_homology_graph(graph_file)

            new_indiset_nodes = possible_nodes - vercov_nodes
            indiset_nodes |= new_indiset_nodes

            possible_nodes = vercov_nodes
            prev_possibility_count = len(possible_nodes)
            for indi_node in new_indiset_nodes:
                possible_nodes.difference_update(homology_graph[indi_node])
            curr_possibility_count = len(possible_nodes)

            if verbose:
                print(' [+] Current independent set size:  {}'.format(len(indiset_nodes)))
                print(' [+] Potential nodes for expansion: {} (projected independent set size: {})'.format(len(possible_nodes), len(indiset_nodes)+len(possible_nodes)))
            if len(possible_nodes) == 0 or prev_possibility_count == curr_possibility_count:
                if verbose:
                    print(' [X] Cannot expand independent set, terminating.')
                break

    return indiset_nodes


from Bio.SeqUtils import MeltingTemp
from time         import time

class Synthesis(object):

    # Setup synthesis assessment variables
    comp_table = str.maketrans('ATGC', 'TACG')
    runs_tuple = (('CCCCCCCCC',            'GGGGGGGGG'),
                  ('AAAAAAAAAAAAA',        'TTTTTTTTTTTTT'),
                  ('TCTTCTTCTTCTTCTTCT',   'TCGTCGTCGTCGTCGTCG',   'TCCTCCTCCTCCTCCTCC',   'GGAGGAGGAGGAGGAGGA',   'GGTGGTGGTGGTGGTGGT',
                   'GGCGGCGGCGGCGGCGGC',   'AATAATAATAATAATAAT',   'AAGAAGAAGAAGAAGAAG',   'AACAACAACAACAACAAC',   'CCACCACCACCACCACCA',
                   'CCTCCTCCTCCTCCTCCT',   'CCGCCGCCGCCGCCGCCG',   'ATAATAATAATAATAATA',   'ATTATTATTATTATTATT',   'ATGATGATGATGATGATG',
                   'ATCATCATCATCATCATC',   'AGAAGAAGAAGAAGAAGA',   'AGTAGTAGTAGTAGTAGT',   'AGGAGGAGGAGGAGGAGG',   'AGCAGCAGCAGCAGCAGC',
                   'ACAACAACAACAACAACA',   'ACTACTACTACTACTACT',   'ACGACGACGACGACGACG',   'ACCACCACCACCACCACC',   'TAATAATAATAATAATAA',
                   'TATTATTATTATTATTAT',   'TAGTAGTAGTAGTAGTAG',   'TACTACTACTACTACTAC',   'TTATTATTATTATTATTA',   'TTGTTGTTGTTGTTGTTG',
                   'TTCTTCTTCTTCTTCTTC',   'TGATGATGATGATGATGA',   'TGTTGTTGTTGTTGTTGT',   'TGGTGGTGGTGGTGGTGG',   'TGCTGCTGCTGCTGCTGC',
                   'TCATCATCATCATCATCA',   'GAAGAAGAAGAAGAAGAA',   'GATGATGATGATGATGAT',   'GAGGAGGAGGAGGAGGAG',   'GACGACGACGACGACGAC',
                   'GTAGTAGTAGTAGTAGTA',   'GTTGTTGTTGTTGTTGTT',   'GTGGTGGTGGTGGTGGTG',   'GTCGTCGTCGTCGTCGTC',   'GCAGCAGCAGCAGCAGCA',
                   'GCTGCTGCTGCTGCTGCT',   'GCGGCGGCGGCGGCGGCG',   'GCCGCCGCCGCCGCCGCC',   'CAACAACAACAACAACAA',   'CATCATCATCATCATCAT',
                   'CAGCAGCAGCAGCAGCAG',   'CACCACCACCACCACCAC',   'CTACTACTACTACTACTA',   'CTTCTTCTTCTTCTTCTT',   'CTGCTGCTGCTGCTGCTG',
                   'CTCCTCCTCCTCCTCCTC',   'CGACGACGACGACGACGA',   'CGTCGTCGTCGTCGTCGT',   'CGGCGGCGGCGGCGGCGG',   'CGCCGCCGCCGCCGCCGC'),
                  ('ATATATATATATATATATAT', 'AGAGAGAGAGAGAGAGAGAG', 'ACACACACACACACACACAC', 'TATATATATATATATATATA', 'TGTGTGTGTGTGTGTGTGTG', 'TCTCTCTCTCTCTCTCTCTC',
                   'GAGAGAGAGAGAGAGAGAGA', 'GTGTGTGTGTGTGTGTGTGT', 'GCGCGCGCGCGCGCGCGCGC', 'CACACACACACACACACACA', 'CTCTCTCTCTCTCTCTCTCT', 'CGCGCGCGCGCGCGCGCGCG'))

    def _is_GC_content_pass(self, seq, start, end, win, gc_low, gc_high):
        """
        >>> synth_obj._is_GC_content_pass(seq='AAAAAAAAAAAAAAAAAAGG', start=0, end=20, win=20, gc_low=0.10, gc_high=0.25)
        False
        >>> synth_obj._is_GC_content_pass(seq='AAAAAAAAAAAAAAAGGGGG', start=0, end=20, win=20, gc_low=0.10, gc_high=0.25)
        False
        >>> synth_obj._is_GC_content_pass(seq='AAAAAAAAAAAAAAAAGGGG', start=0, end=20, win=20, gc_low=0.10, gc_high=0.25)
        True
        """

        gc_count, i = 0.0, start
        while i < end:
            gc_count += seq[i] in ['G', 'C']
            if (gc_count / win) >= gc_high:
                return False
            if i >= (start + win - 1):
                if (gc_count / win) <= gc_low:
                    return False
                gc_count -= seq[i-win+1] in ['G', 'C']
            i += 1
        return True

    def _rule_GC_content(self, seq):
        """
        >>> synth_obj._rule_GC_content(seq='GGGGGGGGGGGGGGAAAAAA')
        False
        >>> synth_obj._rule_GC_content(seq='GGGGGGAAAAAAAAAAAAAA')
        False
        >>> synth_obj._rule_GC_content(seq='GGGGGGGGGGAAAAAAAAAA')
        True
        >>> synth_obj._rule_GC_content(seq='GAAAAAAAAAAAAAAAAAGGGGGGGGGGGG')
        False
        >>> synth_obj._rule_GC_content(seq='GAAAAAAAAAAAAAAAAAGGGGGGGGGGGG')
        False
        >>> synth_obj._rule_GC_content(seq='GAGAGAGAGAGAGAGAGAGAGAGAGAGAGA')
        True
        """

        # Throw out insane sequences
        # if not self._is_GC_content_pass(seq=seq, start=0, end=len(seq), win=len(seq), gc_low=0.30, gc_high=0.68):
        #     return False
        # else:
        # Windowed GC content checkup
        if len(seq) >= 20  and not self._is_GC_content_pass(seq=seq, start=0, end=len(seq), win=20,  gc_low=0.15, gc_high=0.90):
            return False
        if len(seq) >= 100 and not self._is_GC_content_pass(seq=seq, start=0, end=len(seq), win=100, gc_low=0.28, gc_high=0.76):
            return False
        # Terminal GC content checkup
        if len(seq) > 60:
            # 5' end
            if not self._is_GC_content_pass(seq=seq, start=0, end=30, win=30, gc_low=0.24, gc_high=0.76):
                return False
            # 3' end
            if not self._is_GC_content_pass(seq=seq, start=len(seq)-30, end=len(seq), win=30, gc_low=0.24, gc_high=0.76):
                return False
        # All GC conditions satisfy
        return True

    def _is_melting_temp_pass(self, seq, tm_low, tm_high):
        """
        >>> synth_obj._is_melting_temp_pass(seq='AAAAAAAAAATTTTTTTTTT', tm_low=36.0, tm_high=73.0)
        False
        >>> synth_obj._is_melting_temp_pass(seq='GCGCGCGCGCGCATATATAT', tm_low=36.0, tm_high=73.0)
        True
        """

        if tm_low <= MeltingTemp.Tm_NN(seq, dnac1=250.0, dnac2=0.0, saltcorr=7) <= tm_high:
            return True
        return False

    def _rule_melting_temp(self, seq):
        """
        >>> synth_obj._rule_melting_temp(seq='AAAAAAAAAATTTTTTTTTT')
        False
        >>> synth_obj._rule_melting_temp(seq='GCGCGCGCGCGCATATATAT')
        True
        """

        i, win = 0, 20
        if len(seq) >= win:
            while i < len(seq)-win+1:
                if not self._is_melting_temp_pass(seq=seq[i:i+win], tm_low=36.0, tm_high=73.0):
                    return False
                i += 1
        return True

    def _is_slippage_pass(self, seq, start, end, rep_len, win):
        end = min([start+(rep_len*win), end])
        # Direct repeat slippage
        beg = start+1
        pattern = seq[start:start+rep_len]
        if seq.find(pattern, beg, end) > -1:
            return False
        # Inverted repeat slippage, causing a hairpin
        beg = start+rep_len
        pattern = pattern.translate(self.comp_table)[::-1]
        if seq.find(pattern, beg, end) > -1:
            return False
        return True

    def _rule_slippage(self, seq):
        start, rep_len, win, term_len = 0, 5, 3, 60
        # Check terminals of candidate sequence for tandem repeats
        if len(seq) > rep_len:
            # Too small for terminal checks, search fully
            if len(seq) <= term_len:
                start_max, end = len(seq)-rep_len, len(seq)
            # Initiate 5' terminal check
            else:
                start_max, end = term_len-rep_len, term_len
            while start < start_max:
                if not self._is_slippage_pass(seq, start, end, rep_len, win):
                    return False
                start += 1
                # Switch to 3' terminal check
                if start == term_len-rep_len:
                    start, start_max, end = max([start, len(seq) - term_len]), len(seq)-rep_len, len(seq)
        return True

    def _is_runs_pass(self, seq, term_len):
        beg, end = term_len, len(seq)-term_len
        for run_tuple in self.runs_tuple:
            # Non-terminal region long enough to have runs
            if len(run_tuple[0]) <= len(seq)-(2*term_len):
                for run in run_tuple:
                    if seq.find(run, beg, end) > -1:
                        return False
            else:
                break
        return True

    def _rule_runs(self, seq):
        term_len = 60
        # Check for bad runs in non-terminal region only (compliments _rule_slippage)
        return self._is_runs_pass(seq, term_len)

    def _is_hairpin_pass(self, seq, stem, loop, max_mismatch, gc_high):
        i = 0
        while i < len(seq)-(stem+loop+stem)+1:
            j, mismatch, gc_count = 0, 0, 0.0
            while j < stem:
                if seq[i+j] in ['G', 'C']:
                    gc_count += 1.0
                if seq[i+(stem+loop+stem)-j-1] in ['G', 'C']:
                    gc_count += 1.0
                if seq[i+j] != seq[i+(stem+loop+stem)-j-1].translate(self.comp_table):
                    mismatch += 1
                if mismatch > max_mismatch:
                    break
                j += 1
            if j == stem:
                gc_content = (gc_count / 2.0) / stem
                if gc_content >= gc_high:
                    return False
            i += 1
        return True

    def _rule_palindrome(self, seq):
        stem, loop, max_mismatch, gc_high = 11, 0, 1, 0.0
        if len(seq) >= (stem+loop+stem):
            return self._is_hairpin_pass(seq, stem, loop, max_mismatch, gc_high)
        return True

    def _rule_hairpin(self, seq):
        # Type 1: Stem = 11bp, Loop = 3bp to 48bp, Mismatches = 2, Stem must have GC-content >= 0.80
        stem, max_mismatch, gc_high = 11, 2, 0.80
        for loop in range(3, 48+1):
            if len(seq) >= (stem+loop+stem):
                if not self._is_hairpin_pass(seq, stem, loop, max_mismatch, gc_high):
                    return False
            else:
                break

        # Type 2: Stem = 17bp, Loop = 3bp to 100bp, Mismatches = 3
        stem, max_mismatch, gc_high = 17, 3, 0.0
        for loop in range(3, 100+1):
            if len(seq) >= (stem+loop+stem):
                if not self._is_hairpin_pass(seq, stem, loop, max_mismatch, gc_high):
                    return False
            else:
                break

        if len(seq) > 500:
            # Type 3: Stem = 22bp, Loop = 100bp to 500bp, Mismatches = 5
            stem, max_mismatch, gc_high = 22, 5, 0.0
            for loop in range(100, 500+1):
                if len(seq) >= (stem+loop+stem):
                    if not self._is_hairpin_pass(seq, stem, loop, max_mismatch, gc_high):
                        return False
                else:
                    break

        return True

    def evaluate_short(self, seq):
        if self._rule_GC_content(seq):
            if self._rule_runs(seq):
                if self._rule_palindrome(seq):
                    if self._rule_hairpin(seq):
                        return True
        return False

    def evaluate_long(self, seq):
        if self._rule_GC_content(seq):
            if self._rule_slippage(seq):
                if self._rule_runs(seq):
                    if self._rule_palindrome(seq):
                        if self._rule_melting_temp(seq):
                            if self._rule_hairpin(seq):
                                return True
        return False

    def evaluate(self, seq):
        if len(seq) <= 500:
            return self.evaluate_short(seq)
        else:
            return self.evaluate_long(seq)



if __name__ == '__main__':
    import doctest
    doctest.testmod(extraglobs={'synth_obj': Synthesis()})

    seq = 'GGGCAGATGTCACTCAGTAGGATTTGATCAGTACGTGCATGATGCTGCGGGCAGATGTCACTCAGTAGGATTTGATCAGTACGTGCATGATGCTGC'*100#*1000
    # seq = 'AGTCGCAGACAAAAAAAAAAATTTTTTTTTTTCGATGCTAGTCG'

    # seq = 'GCGC GACG ATGA CCGG CAGA CCGG TCAT CGTC ATAT'
    # seq = 'GCGCGACGATGACCGGCAGACCGGTCATCGTCATAT'

    # synth_obj = synthesis()
    # t0 = time()
    # print synth_obj.evaluate(seq)
    # print 'Time elapsed: {}'.format(time()-t0)


    # t0 = time()
    # # print synth_obj._rule_GC_content(seq)
    # # print synth_obj._rule_melting_temp(seq)
    # # print synth_obj._rule_slippage(seq)
    # # print synth_obj._rule_runs(seq)
    # # print synth_obj._rule_palindrome(seq)
    # print synth_obj._rule_hairpin(seq)
    # print 'Time elapsed: {}'.format(time()-t0)


from .base import maker     as nrpmaker
from .base import finder    as nrpfinder
from .base import kmerSetDB


__version__ = '1.6.3'

__authors__ = '''
Ayaan Hossain <auh57@psu.edu>
Howard Salis  <salis@psu.edu>
'''

__doc__ = '''
Non-Repetitive Parts Calculator

Automated design and discovery of non-repetitive genetic
parts for engineering stable systems.

Version: {}

Authors: {}
         {}

The Non-Repetitive Parts Calculator offers two modes of operation:

- Finder Mode: Discover toolboxes of non-repetitive parts
               from a list of candidate parts

-  Maker Mode: Design toolboxes of non-repetitive parts
               based on sequence, structure and model
               constraints

Additionally, a 'background' object is available that stores
background sequences. When the 'background' object is used, 
designed genetic parts will also be non-repetitive with respect
to these sequences.

You can learn more about the two modes and background via
  print(nrpcalc.background.__doc__)
  print(nrpcalc.finder.__doc__)
  print(nrpcalc.maker.__doc__)
'''.format(
    __version__,
    *__authors__.strip().split('\n'))


def background(
    path,
    Lmax,
    verbose=True):
    '''
    NRP Calculator kmerSetDB background object for on-disk
    storage of background sequence k-mers (where k=Lmax+1).
    When a sequence is added to background, k-mers from the
    sequence are added instead of the actual sequence itself.
    A sequence queried for existence in the given background
    is evaluated to be True if any k-mer from the sequence
    exists in the background. This object is useful when
    chaining multiple Maker Mode and Finder Mode jobs as
    well as persisting any backgrounds such as small genomes
    or other part toolboxes from earlier design rounds.

    :: path
       type - string
       desc - ./a/path/to/store/background/object for later
              reuse and part verification
    :: Lmax
       type - integer
       desc - maximum allowed shared repeat length between
              all sequences in a given toolbox
    :: verbose
       type - boolean
       desc - if True displays progress
              (default=True)

    Returns: A kmerSetDB object.

    Note: If the path provided points to an existing background
          object, then that background is opened for reading, and
          and the new Lmax is ignored, otherwise, a new background
          is instantiated at the given path.

    background / kmerSetDB API Examples
    
    >>> from pprint import pprint
    >>>
    >>> import nrpcalc
    >>>
    >>> my_background_list = [
        'ATGAGATCGTAGCAACC',
        'GACGATTACGTCAGGTA',
        'ACAGTAGAGACGAGTAA',
        'CCAGTACGAAAAGGCCC',
        'TTAGCTTGATAGTTTTA']
    >>> bkg = nrpcalc.background(
            path='./prj_bkg/',
            Lmax=15)
    >>> bkg
    kmerSetDB stored at ./prj_bkg/ with 0 16-mers
    >>>    

    background / kmerSetDB offers the following methods

    (1) add(seq) - adds an IUPAC string 'seq' to background

    >>> bkg.add('ATGCTTAGTGCCATACC')
    
    (2) multiadd(seq_list) - adds multiple sequences from
                             the list to background

    >>> bkg.multiadd(my_background_list)
    
    [Background Processing]
      Adding Seq 4: TTAGCTTGAT...
    
    (3) __contains__(seq) - checks if all k-mers from seq
                            is present in background

    >>> 'ATGCTTAGTGCCATACC' in bkg
    True

    (4) multicheck(seq_list) - checks if all k-mers from given
                               seq_list present in background

    >>> assert all(bkg.multicheck(my_background_list))

    (5) __iter__() - iterate over all k-mers in background

    >>> pprint(list(bkg))
    ['AAAACTATCAAGCTAA',
     'ACAGTAGAGACGAGTA',
     'ACCTGACGTAATCGTC',
     'ACGATTACGTCAGGTA',
     'ATGAGATCGTAGCAAC',
     'ATGCTTAGTGCCATAC',
     'CAGTACGAAAAGGCCC',
     'CAGTAGAGACGAGTAA',
     'CCAGTACGAAAAGGCC',
     'GGTATGGCACTAAGCA',
     'GGTTGCTACGATCTCA',
     'TAAAACTATCAAGCTA']

    (6) __len__() - returns the number of k-mers in background

    >>> len(bkg)
    12

    (7) remove(seq) - removes all k-mers in seq from the
                      background, freeing them up for use

    >>> 'ATGCTTAGTGCCATACC' in bkg
    True
    >>> bkg.remove('ATGCTTAGTGCCATACC')
    >>> 'ATGCTTAGTGCCATACC' in bkg
    False
    >>> len(bkg)
    10

    (8) multiremove(seq_list) - removes all k-mers from given
                                seq_list from background

    >>> bkg.multiremove(my_background_list)

    [Background Processing]
      Removing Seq 4: TTAGCTTGAT
    >>> len(bkg)
    0

    (9) clear(Lmax=None) - removes all k-mers stored in background,
                           and optionally resets background Lmax

    >>> bkg.add('ATGCTTAGTGCCATACC')
    >>> len(bkg)
    2
    >>> bkg.clear()
    >>> len(bkg)
    0
    >>> bkg
    kmerSetDB stored at ./prj_bkg/ with 0 16-mers
    >>> bkg.clear(Lmax=17)
    >>> bkg
    kmerSetDB stored at ./prj_bkg/ with 0 18-mers
    >>> bkg.clear(Lmax=15)
    >>> bkg
    kmerSetDB stored at ./prj_bkg/ with 0 16-mers

    (10) close() - closes background instance; once closed
                   operations on the instance raises error
    
    >>> bkg.close()
    True
    >>> bkg.close()
    False
    >>> bkg.add('ATGCTTAGTGCCATACC')
    Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
      File "nrpcalc/base/kmerSetDB.py", line 97, in wrapper
        raise RuntimeError('kmerSetDB was dropped')
    RuntimeError: kmerSetDB was closed or dropped

    (11) drop() - deletes unclosed `background` from disk;
                  once dropped operations on the instance
                  raises error

    >>> bkg.drop()
    False
    >>> bkg = nrpcalc.background(
            path='./prj_bkg/',
            Lmax=15)
    >>> bkg
    kmerSetDB stored at ./prj_bkg/ with 0 16-mers
    >>> bkg.drop()
    True
    >>> bkg.add('ATGCTTAGTGCCATACC')
    Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
      File "nrpcalc/base/kmerSetDB.py", line 97, in wrapper
        raise RuntimeError('kmerSetDB was closed or dropped')
    RuntimeError: kmerSetDB was closed or dropped
    '''
    return kmerSetDB.kmerSetDB(
        path=path,
        homology=Lmax+1,
        verbose=verbose)

def finder(
    seq_list,
    Lmax,
    internal_repeats=False,
    background=None,
    vercov='nrp2',
    output_file=None,
    verbose=True):
    '''
    NRP Calculator Finder Mode for discovering non-repetitive
    subset of parts from a given list. All parts sharing any
    repeat longer than Lmax are eliminated from seq_list, and
    the approximately largest subset of non-repetitive parts
    is returned in a dictionary indexed by their position in
    seq_list. If internal_repeats is set to True, then parts
    with internal repeats are preserved, otherwise such parts
    are eliminated from seq_list. Optionally, the discovered
    subset of parts is written to an output FASTA file.

    :: seq_list
       type - list
       desc - a list of IUPAC strings representing a genetic
              part toolbox
    :: Lmax
       type - integer
       desc - maximum allowed shared repeat length between
              all sequences in a given toolbox
    :: internal_repeats
       type - boolean
       desc - if False then parts containing internal repeats
              longer than Lmax are eliminated; shared repeats
              are eliminated regardless
              (default=False)
    :: background
       type - kmerSetDB / None
       desc - the background object containing k-mers (k=Lmax+1)
              which must be absent in discovered non-repetitive
              subset of parts
              (default=None)
    :: vercov
       type - string
       desc - must be either '2apx', 'nrpG', or 'nrp2'
              '2apx' - use standard 2-approximation Vertex
                       Cover Elimination algorithm
              'nrpG' - use Greedy Vertex Cover Elimination
                       algorithm
              'nrp2' - use Finder Mode 2-approximation Vertex
                       Cover Elimination algorithm
              (default='nrp2')
    :: output_file
       type - string / None
       desc - filename to store discovered non-repetitive parts
              indexed by their position in seq_list; sequences
              are written in FASTA format
              (default=None)
    :: verbose
       type - boolean
       desc - if True displays progress
              (default=True)

    Returns: A dictionary of DNA or RNA strings with integer keys.

    Finder Mode API Example

    >>> import nrpcalc
    >>> 
    >>> # define background corpus
    >>> my_chromosomes = [
        'ATGAGATCGTAGCAACC',
        'GACGATTACGTCAGGTA',
        'ACAGTAGAGACGAGTAA',
        'CCAGTACGAAAAGGCCC',
        'AAAAAAAAAAAAAAAAA']
    >>> 
    >>> # initialize background
    >>> genomic_kmers = nrpcalc.background(
        path='./my_genome/',
        Lmax=15)
    >>> genomic_kmers.multiadd(
        my_chromosomes)

    [Background Processing]
      Adding Seq 4: AAAAAAAAAA...
    >>> 
    >>> # fetch part toolbox
    >>> my_toolbox = [
        'AGAGCTATGACTGACGT',
        'GCAGATAGGGGGTAGTA',
        'TAAAAAAAAAAAAAAAA', # Repeats with last chromosome
        'GAGCTATGACTGACGTC'] # Repeats with first part
    >>> 
    >>> # find non-repetitive subset
    >>> # with respect to background
    >>> nrpcalc.finder(
        seq_list=my_toolbox,
        Lmax=15,
        background=genomic_kmers)

    [Non-Repetitive Parts Calculator - Finder Mode]

    [Checking Constraints]
     Sequence List   : 4 parts
              Lmax   : 15 bp
     Internal Repeats: False

     Check Status: PASS

    [Checking Background]
     Background: kmerSetDB stored at ./my_genome/ with 10 16-mers

     Check Status: PASS

    [Checking Arguments]
       Vertex Cover: nrp2
       Output  File: None

     Check Status: PASS

    Extracted 4 unique sequences out of 4 sequences in 3.91e-05 seconds

    Written 4 unique sequences out to ./30c21235-e3f6-47f5-bce1-f99f47053e0b/seq_list.txt in 0.0002167 seconds

     [Sequence processing remaining] = 1 
     [Cliques inserted] = 2 

    Built homology graph in 0.0006673 seconds. [Edges = 1] [Nodes = 3]
     [Intital Nodes = 4] - [Repetitive Nodes = 1] = [Final Nodes = 3]

     [+] Initial independent set = 0, computing vertex cover on remaining 0 nodes.
     [+] Vertex Cover Function: NRP 2-approximation
     [+] Dumping graph into: ./30c21235-e3f6-47f5-bce1-f99f47053e0b/repeat_graph.txt in 0.00023055076599121094 seconds

    ----------------------
    Now running iteration: 0
    ----------------------

     Pendant checking is in progress...
      [+] 3 Pendants found

     Pendant elimination initiated...
      [x] Isolated node 1 eliminated
      [x] Pendant node 0 eliminated
      [+] Node 3 covered


     [+] Computed vertex cover of size: 1 (in 0.0002351 seconds)
     [+] Loading graph from: ./30c21235-e3f6-47f5-bce1-f99f47053e0b/repeat_graph.txt
     [+] Current independent set size:  2
     [+] Potential nodes for expansion: 0 (projected independent set size: 2)
     [X] Cannot expand independent set, terminating.

    Non-Repetitive Toolbox Size: 2
    {1: 'GCAGATAGGGGGTAGTA', 0: 'AGAGCTATGACTGACGT'}
    >>>
    >>> genomic_kmers.drop() # we're done with this background
    True
    '''
    return nrpfinder.nrp_finder(
        seq_list=seq_list,
        homology=Lmax+1,
        background=background,        
        allow_internal_repeat=internal_repeats,
        vercov_func=vercov,
        output_file=output_file,
        verbose=verbose,
        check_constraints=True)

def maker(
    seq_constr,
    struct_constr,
    part_type,
    Lmax,
    target_size,
    internal_repeats=False,
    background=None,
    struct_type='mfe',
    seed=None,
    synth_opt=False,
    local_model_fn=None,
    global_model_fn=None,
    jump_count=10,
    fail_count=1000,
    output_file=None,
    verbose=True):
    '''
    NRP Calculator Maker Mode for designing non-repetitive
    genetic part toolboxes from user defined sequence and
    structure constraints and based on custom local and/or
    global model functions. All shared repeats longer than
    Lmax are eliminated, and internal repeats longer than
    Lmax are preserved if desired. Parts are optimized for
    DNA synthesis if desired. Error tolerance is adaptive
    and auto-adjusted based on recorded failures. Designed
    toolbox is returned as a dictionary of parts indexed by
    their order of design, and optionally written to a FASTA
    output file.

    :: seq_constr
       type - string
       desc - a string in IUPAC degenerate code describing
              all valid nucleotide choices at each position
              e.g. 'NNNNWWWWSSSSTTTT' implies that the first
                   four bases can be either 'A'/'T'/'G'/'C',
                   the next four bases can be either 'A'/'T',
                   followed by either 'G'/'C' for the next
                   four basses, and finally ending with 'T's
    :: struct_constr
       type - string / None
       desc - a string in dot-parenthesis-x notation that
              describe the secondary base pairing across all
              nucleotide positions; a string of only dots implies
              an absence of any structure constraint
              e.g. '..((xx))..' implies that the first, second,
                   and the last two bases are free to either
                   base pair or not ('.'), the third and fourth
                   bases are paired with the eighth and the
                   seventh bases respectively ('(' and ')'),
                   while the fifth and the sixth base must
                   not take part in any base pairing ('x') at all
    :: part_type
       type - string
       desc - must be either 'RNA' or 'DNA' depending on the
              type of genetic part being designed; ensures
              that correct folding free-energy parameters are
              used during structure evaluation
    :: Lmax
       type - integer
       desc - maximum allowed shared repeat length between
              all sequences in designed toolbox
    :: target_size
       type - integer
       desc - maximum number of genetic parts to be designed
              for the generated toolbox; target_size may not
              be reached if the constraints are too strict,
              for example, due to low degeneracy in the given
              sequence constraint, or a low Lmax
    :: internal_repeats
       type - boolean
       desc - if True then internal repeats in designed parts
              are not eliminated; useful when designing parts
              such as rho-independent terminators with structure
              constraints that necessitate internal repeats;
              shared repeats are eliminated regardless
              (default=False)
    :: background
       type - kmerSetDB / None
       desc - a background object containing k-mers (k=Lmax+1)
              which must be absent in the designed toolbox
              (default=None)
    :: struct_type
       type - string
       desc - must be either 'mfe', 'centroid', or 'both'
              'mfe' - use minimum free energy structure evaluation
              'centroid' - use centroid structure evaluation
              'both' - use both 'mfe' and 'centroid' evaluation
              (default='mfe')
    :: seed
       type - integer / None
       desc - integer used to seed random number generations;
              two Maker runs with same constraints and seed
              value will generate the exact same toolbox;
              if None then a random seed value is used
              (default=None)
    :: synth_opt
       type - boolean
       desc - if True then designed parts containing features
              that complicate DNA synthesis are eliminated
              (default=False)
    :: local_model_fn
       type - function / None
       desc - a function with signature 'fn_name(seq)' that
              takes in a partial genetic part sequence, and
              returns a tuple (state, index), where state is
              either True or False and index is a traceback
              index / location or None depending on whether
              a custom design objective was met or not;
              useful for providing concurrent feedback to the
              path-finding process and steering nucleotide
              selection choices
              e.g. prevent_cutsites(seq) maybe be a local
                   function that takes in a partial sequence
                   as it is built and returns (True, None)
                   if the last six bases of the partial
                   seq[-6:] is not the same as any of the
                   cutsites used for cloning the part, else
                   returns a tuple (False, len(seq)-6) as
                   the traceback location to reselect the
                   last six bases; this naturally ensures
                   the final part is devoid of any cutsites
                   used experimentally throughout the part
              (default=None)
    :: global_model_fn
       type - function / None
       desc - a function with signature 'fn_name(seq)' that
              takes in a complete genetic part sequence, and
              returns either True or False depending on whether
              a custom design objective was met; useful for
              design criteria that can only be evaluated when
              the complete genetic part is available; parts
              that are evaluated to be False are rejected and
              a new part generation is started; global model
              functions are evaluated only after the last base
              has been added to a genetic part under design
              e.g. gc_content(seq) may be a global function
                   that takes in a complete sequence and only
                   accepts parts with GC content greater than
                   threshold percentage (although, technically
                   one can enforce this condition via a well
                   planned local model function)
              (default=None)
    :: jump_count
       type - integer
       desc - maximum number of restarts in path finding due
              to failure in finding suitable k-mers, meeting
              local_model_fn, or being stuck in local optima
              (auto-adjusted with each iteration)
              (default=10)
    :: fail_count
       type - integer
       desc - maximum number of consecutive failures tolerated
              when structure constraints, global model functions
              and synthesis objectives are not met
              (auto-adjusted with each iteration)
              (default=1000)
    :: output_file
       type - string / None
       desc - filename to store designed non-repetitive parts
              as they are generated consecutively; sequences
              are written in FASTA format
              (default=None)
    :: verbose
       type - boolean
       desc - if True displays progress
              (default=True)

    Returns: A dictionary of DNA or RNA strings with integer keys.
    
    Maker Mode API Example
    
    >>> import nrpcalc
    >>>
    >>> Lmax = 15  # for global use
    >>>
    >>> # initialize background
    >>> bkg = nrpcalc.background(
        path='./my_toolbox_kmers',
        Lmax=Lmax)
    >>> bkg
    kmerSetDB stored at ./my_toolbox_kmers/ with 0 16-mers
    >>>
    >>> # a local model function
    >>> def prevent_cutsites(seq):
            # part is long enough to evaluate
            if len(seq) >= 6:
                BamHI = 'GGATCC' # cutsite 1
                XbaI  = 'TCTAGA' # cutsite 2
                # evaluate part concurrently
                if seq[-6:] in [BamHI, XbaI]:
                    return (False, len(seq)-6)
                else:
                    return (True, None)
            # part is short enough .. pass
            return (True, None)
    >>>
    >>> # a global model function
    >>> def optimal_gc_content(seq):
            # compute gc count
            gcount = seq.count('G') * 1.
            ccount = seq.count('C') * 1.
            # evaluate on complete part
            if 0.4 <= (gcount + ccount) / len(seq) <= 0.6:
                return True
            else:
                return False
    >>>
    >>> # final toolbox storage
    >>> final_toolbox = []
    >>>
    >>> # build a toolbox of non-repetitive sigma70 promoters
    >>> promoters_strong = nrpcalc.maker(
        seq_constr='N'*20+'TTGACA'+'N'*17+'TATAAT'+'NNNNNN',
        struct_constr='.'*55,
        part_type='DNA',
        Lmax=Lmax,
        target_size=500,
        internal_repeats=False,
        background=None,
        local_model_fn=prevent_cutsites,
        global_model_fn=optimal_gc_content)
    WARNING: stacking enthalpies not symmetric
    WARNING: stacking enthalpies not symmetric
    WARNING: stacking enthalpies not symmetric
    WARNING: stacking enthalpies not symmetric

    [Non-Repetitive Parts Calculator - Maker Mode]

    [Checking Constraints]
      Sequence Constraint: NNNNNNNNNNNNNNNNNNNNTTGACANNNNNNNNNNNNNNNNNTATAATNNNNNN
     Structure Constraint: .......................................................
          Part Type      : DNA
               Lmax      : 15 bp
        Target Size      : 500 parts
      Internal Repeats   : False

     Check Status: PASS

    [Checking Arguments]
     Struct Type : mfe
      Synth Opt  : False
       Jump Count: 10
       Fail Count: 1000
     Output File : None

     Check Status: PASS

    Constructing Toolbox:

     [part] 1, [16-mers] 40, [iter 1] 0.00s, [avg] 0.00s, [total time] 0.00h
     [part] 2, [16-mers] 80, [iter 2] 0.01s, [avg] 0.01s, [total time] 0.00h
     ...
     [part] 499, [16-mers] 19960, [iter 499] 0.00s, [avg] 0.00s, [total time] 0.00h
     [part] 500, [16-mers] 20000, [iter 500] 0.00s, [avg] 0.00s, [total time] 0.00h

    Construction Complete.

    Non-Repetitive Toolbox Size: 500
    >>>
    >>> # add promoter toolbox to background
    >>> bkg.multiadd(promoters_strong.values())

    [Background Processing]
      Adding Seq 499: AACATCGATG... 
    >>>
    >>> # add promoters to final_toolbox
    >>> final_toolbox.extend(promoters_strong.values())
    >>>
    >>> # build another toolbox of sigma70 promoters
    >>> # non-repetitive to the previous toolbox
    >>> # (this is called toolbox chaining)
    >>> promoters_variable = nrpcalc.maker(
        seq_constr='N'*20+'TTGACA'+'N'*16+'WWWWWWW'+'NNNNN',
        struct_constr='.'*55,
        part_type='DNA',
        Lmax=Lmax,
        target_size=500,
        internal_repeats=False,
        background=bkg,
        local_model_fn=prevent_cutsites,
        global_model_fn=optimal_gc_content)
    WARNING: stacking enthalpies not symmetric
    WARNING: stacking enthalpies not symmetric
    WARNING: stacking enthalpies not symmetric
    WARNING: stacking enthalpies not symmetric

    [Non-Repetitive Parts Calculator - Maker Mode]

    [Checking Constraints]
      Sequence Constraint: NNNNNNNNNNNNNNNNNNNNTTGACANNNNNNNNNNNNNNNNWWWWWWWNNNNN
     Structure Constraint: .......................................................
        Target Size      : 500 parts
               Lmax      : 15 bp
      Internal Repeats   : False

     Check Status: PASS

    [Checking Background]
     Background: kmerSetDB stored at ./my_toolbox_kmers/ with 20000 16-mers

     Check Status: PASS

    [Checking Arguments]
       Part Type : DNA
     Struct Type : mfe
      Synth Opt  : False
       Jump Count: 10
       Fail Count: 1000
     Output File : None

     Check Status: PASS

    Constructing Toolbox:

     [part] 1, [16-mers] 39, [iter 1] 0.01s, [avg] 0.01s, [total time] 0.00h
     [part] 2, [16-mers] 78, [iter 2] 0.00s, [avg] 0.01s, [total time] 0.00h
     ...
     [part] 499, [16-mers] 19461, [iter 499] 0.00s, [avg] 0.00s, [total time] 0.00h
     [part] 500, [16-mers] 19500, [iter 500] 0.00s, [avg] 0.00s, [total time] 0.00h

    Construction Complete.

    Non-Repetitive Toolbox Size: 500
    >>>
    >>> # remove background .. it's served its purpose
    >>> bkg.drop()
    True
    >>>
    >>> # add new promoters to final_toolbox
    >>> final_toolbox.extend(promoters_variable.values())
    >>>
    >>> # verify all promoters are non-repetitive
    >>> assert len(nrpcalc.finder(
        seq_list=final_toolbox,
        Lmax=Lmax,
        verbose=False)) == 1000
    '''
    # Initialize object
    _maker = nrpmaker.NRPMaker(
        part_type=part_type,
        seed=seed)
    
    # Compute parts
    parts = _maker.nrp_maker(
        seq_constr=seq_constr,
        struct_constr=struct_constr,
        struct_type=struct_type,
        target_size=target_size,
        homology=Lmax+1,
        allow_internal_repeat=internal_repeats,
        background=background,
        synth_opt=synth_opt,
        local_model_fn=local_model_fn,
        global_model_fn=global_model_fn,
        jump_count=jump_count,
        fail_count=fail_count,
        output_file=output_file,
        verbose=verbose,
        abortion=True,
        check_constraints=True)

    # Cleanups and return
    del _maker
    return parts